# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable disciplines from Tedford and Martins (2010)
****************************************************
"""
from __future__ import division, unicode_literals

import logging

from gemseo.core.discipline import MDODiscipline
from gemseo.problems.scalable.parametric.core.models import TMMainModel, TMSubModel
from gemseo.problems.scalable.parametric.core.variables import (
    X_SHARED_NAME,
    get_coupling_name,
    get_u_local_name,
    get_x_local_name,
)

LOGGER = logging.getLogger(__name__)


class TMDiscipline(MDODiscipline):
    """Abstract base class for disciplines in the TM problem."""

    @property
    def inputs_sizes(self):
        """Sizes of the model inputs."""
        return self.model.inputs_sizes

    @property
    def outputs_sizes(self):
        """Sizes of the model outputs."""
        return self.model.outputs_sizes

    @property
    def inputs_names(self):
        """Names of the model inputs."""
        return self.model.inputs_names

    @property
    def outputs_names(self):
        """Names of the model outputs."""
        return self.model.outputs_names


class TMMainDiscipline(TMDiscipline):

    r"""The system discipline from the scalable problem introduced by Tedford
    and Martins (2010) takes the  local design parameters
    :math:`x_1,x_2,\ldots,x_N` and the global design parameters :math:`z`
    as inputs, as well as the coupling variables :math:`y_1,y_2,\ldots,y_N`
    and returns the objective function value :math:`f(x,y(x,y))` to minimize as
    well as the inequality constraints ones
    :math:`c_1(y_1),c_2(y_2),\ldots,c_N(y_N)` which are expressed as:

    .. math::

       f(z,y) = |z|_2^2 + \sum_{i=1}^N |y_i|_2^2


    and:

    .. math::

       c_i(y_i) = 1- C_i^{-T}Iy_i

    """

    def __init__(self, c_constraint, default_inputs):
        """Constructor.

        :param list(array) c_constraint: constraint coefficients
        :param dict default_inputs: default inputs
        """
        self.model = TMMainModel(c_constraint, default_inputs)
        super(TMMainDiscipline, self).__init__(self.model.name)
        self.input_grammar.initialize_from_data_names(self.model.inputs_names)
        self.output_grammar.initialize_from_data_names(self.model.outputs_names)
        self.default_inputs = default_inputs
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    @property
    def n_disciplines(self):
        """Return the number of disciplines; alias for self.models.n_submodels."""
        return self.model.n_submodels

    def _run(self):
        x_shared = self.get_inputs_by_name(X_SHARED_NAME)
        coupling = {
            get_coupling_name(index): self.get_inputs_by_name(get_coupling_name(index))
            for index in range(self.n_disciplines)
        }
        self.store_local_data(**self.model(x_shared, coupling))

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Computes the jacobian.

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        self._init_jacobian(inputs, outputs, with_zeros=True)
        x_shared = self.get_inputs_by_name(X_SHARED_NAME)
        coupling = {
            get_coupling_name(index): self.get_inputs_by_name(get_coupling_name(index))
            for index in range(self.n_disciplines)
        }
        jac = self.model(x_shared, coupling, jacobian=True)
        for output in jac:
            for inpt in jac[output]:
                self.jac[output][inpt] = jac[output][inpt]


class TMSubDiscipline(TMDiscipline):

    r"""An elementary discipline from the scalable problem introduced by
    Tedford and Martins (2010) takes local design parameters :math:`x_i`
    and shared design parameters :math:`z` in input as well as coupling
    variables :math:`\left(y_i\right)_{1\leq j \leq N\atop j\neq i}`
    from :math:`N-1` elementary disciplines,
    and returns the coupling variables:

    .. math::

        y_i =\frac{\tilde{y}_i+C_{z,i}.1+C_{x_i}.1}{\sum_{j=1 \atop j
        \neq i}^NC_{y_j,i}.1+C_{z,i}.1+C_{x_i}.1} \in [0,1]^{n_{y_i}}

    where:

    .. math::

        \tilde{y}_i = - C_{z,i}.z - C_{x_i}.x_i +
        \sum_{j=1 \atop j \neq i}^N C_{y_j,i}.y_j
    """

    def __init__(self, index, c_shared, c_local, c_coupling, default_inputs):
        """Constructor.

        :param int index: discipline index for naming.
        :param array c_shared: weights for the shared design parameters.
        :param array c_local: weights for the local design parameters.
        :param dict(array) c_coupling: weights for the coupling parameters.
        :param dict default_inputs: default inputs
        """
        self.model = TMSubModel(index, c_shared, c_local, c_coupling, default_inputs)
        super(TMSubDiscipline, self).__init__(name=self.model.name)
        self.input_grammar.initialize_from_data_names(self.model.inputs_names)
        self.output_grammar.initialize_from_data_names(self.model.outputs_names)
        self.default_inputs = default_inputs
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _run(self):
        x_shared = self.get_inputs_by_name(X_SHARED_NAME)
        x_local = self.get_inputs_by_name(get_x_local_name(self.model.index))
        u_local_name = get_u_local_name(self.model.index)
        if self.input_grammar.is_data_name_existing(u_local_name):
            u_local = self.get_inputs_by_name(u_local_name)
        else:
            u_local = None
        coupling = {
            cpl_name: self.get_inputs_by_name(cpl_name)
            for cpl_name in list(self.model.c_coupling.keys())
        }
        self.store_local_data(**self.model(x_shared, x_local, coupling, u_local))

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Computes the jacobian.

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        self._init_jacobian(inputs, outputs, with_zeros=True)
        x_shared = self.get_inputs_by_name(X_SHARED_NAME)
        x_local = self.get_inputs_by_name(get_x_local_name(self.model.index))
        coupling = {
            cpl_name: self.get_inputs_by_name(cpl_name)
            for cpl_name in list(self.model.c_coupling.keys())
        }
        jac = self.model(x_shared, x_local, coupling, jacobian=True)
        for output in jac:
            for inpt in jac[output]:
                self.jac[output][inpt] = jac[output][inpt]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Generate a gantt chart with processes execution time data."""

from __future__ import unicode_literals

from typing import Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt

from gemseo.core.discipline import MDODiscipline
from gemseo.utils.file_path_manager import FilePathManager, FileType
from gemseo.utils.matplotlib_figure import save_show_figure
from gemseo.utils.py23_compat import Path

DEFAULT_NAME = "gantt_chart"


def create_gantt_chart(
    file_path=DEFAULT_NAME,  # type: Union[str, Path]
    save=True,  # type: bool
    show=False,  # type: bool
    file_extension=None,  # type: Optional[str]
    figure_size=(15, 10),  # type: Tuple[int,int]
    font_size=12,  # type: int
    disc_names=None,  # type: Optional[Sequence[str]]
):  # type: (...) -> plt.Figure
    """Generate a gantt chart with processes execution time data.

    The disciplines names are used as labels and plotted on rows.
    The x labels are the execution start and end times of all disciplines.

    Both executions and linearizations times are plotted.

    Warning:
        ``MDODiscipline.activate_time_stamps()`` must be called first.

    Args:
        file_path: The path to save the figure.
        save: Whether to save the figure.
        show: Whether to show the figure.
        file_extension: A file extension, e.g. 'png', 'pdf', 'svg', ...
            If ``None``, use the default file extension.
        figure_size: The figure size.
        font_size: The size of the fonts in the plot.
        disc_names: The names of the disciplines to plot.
            If ``None``, plot all the disciplines for which time stamps exist.

    Returns:
        The matplotlib figure.

    Raises:
        ValueError: If the time stamps is not activated or if a discipline has no
            time stamps.
    """
    time_stamps = MDODiscipline.time_stamps
    if time_stamps is None:
        raise ValueError("Time stamps are not activated in MDODiscipline")

    fig, ax = plt.subplots(figsize=figure_size)

    if disc_names is None:
        disc_names = list(time_stamps.keys())
    else:
        missing = list(set(disc_names) - set(time_stamps.keys()))
        if missing:
            raise ValueError(
                "The disciplines: {}, have no time stamps!".format(missing)
            )

    ax.set_ylim(5, 10 * len(disc_names) + 15)
    ax.set_yticklabels(disc_names)
    ax.set_yticks([5 + (i + 1) * 10 for i in range(len(disc_names))])
    ax.set_xlabel("Time")
    ax.set_ylabel("Disciplines")
    ax.grid(True)

    # Minimum time as a reference
    min_t = min((stamps[0][0] for stamps in time_stamps.values()))

    # Blue for execution, red for linearization
    colors = {False: "tab:blue", True: "tab:red"}
    for i, name in enumerate(disc_names):
        stamps_orig = time_stamps[name]
        stamps = [(l - min_t, u - l) for (l, u, _) in stamps_orig]
        facecolors = [colors[s[2]] for s in stamps_orig]
        ax.broken_barh(stamps, ((i + 1) * 10, 9), facecolors=facecolors)

    # Set all fonts sizes
    for item in (
        [ax.title, ax.xaxis.label, ax.yaxis.label]
        + ax.get_xticklabels()
        + ax.get_yticklabels()
    ):
        item.set_fontsize(font_size)

    if save:
        file_path = FilePathManager(
            FileType.FIGURE, default_name=DEFAULT_NAME
        ).create_file_path(file_path=file_path, file_extension=file_extension)
    else:
        file_path = None

    save_show_figure(fig, show, file_path)

    return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Here is the baseclass to measure the quality of machine learning algorithms.

The concept of clustering quality measure is implemented with the
:class:`.MLClusteringMeasure` class and proposes different evaluation methods.
"""
from __future__ import division, unicode_literals

from typing import Dict, List, NoReturn, Optional, Union

from numpy import arange, array_split
from numpy import delete as npdelete
from numpy import ndarray, unique
from numpy.random import choice

from gemseo.core.dataset import Dataset
from gemseo.mlearning.cluster.cluster import (
    MLClusteringAlgo,
    MLPredictiveClusteringAlgo,
)
from gemseo.mlearning.qual_measure.quality_measure import MLQualityMeasure


class MLClusteringMeasure(MLQualityMeasure):
    """An abstract clustering measure for clustering algorithms."""

    def __init__(
        self,
        algo,  # type: MLClusteringAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm for clustering.
        """
        super(MLClusteringMeasure, self).__init__(algo)

    def evaluate_learn(
        self,
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        if not self.algo.is_trained:
            self.algo.learn(samples)
        data = self._get_data()[samples]
        labels = self.algo.labels
        measure = self._compute_measure(data, labels, multioutput)
        return measure

    def _compute_measure(
        self,
        data,  # type: ndarray
        labels,  # type: ndarray
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        """Compute the quality measure.

        Args:
            data: The reference data.
            labels: The predicted labels.
            multioutput: If True, return the quality measure for each
                output component. Otherwise, average these measures.

        Returns:
            The value of the quality measure.
        """
        raise NotImplementedError

    def _get_data(self):  # type: (...) -> Dict[str,ndarray]
        """Get data.

        Returns:
            The learning data indexed by the names of the variables.
        """
        names = self.algo.var_names
        data = self.algo.learning_set.get_data_by_names(names, False)
        return data


class MLPredictiveClusteringMeasure(MLClusteringMeasure):
    """An abstract clustering measure for predictive clustering algorithms."""

    def __init__(
        self,
        algo,  # type: MLPredictiveClusteringAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm for predictive clustering.
        """
        super(MLPredictiveClusteringMeasure, self).__init__(algo)

    def evaluate_test(
        self,
        test_data,  # type:Dataset
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        if not self.algo.is_trained:
            self.algo.learn(samples)
        names = self.algo.var_names
        data = test_data.get_data_by_names(names, False)
        predictions = self.algo.predict(data)
        measure = self._compute_measure(data, predictions, multioutput)
        return measure

    def evaluate_kfolds(
        self,
        n_folds=5,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        indices = self._assure_samples(samples)
        folds = array_split(indices, n_folds)

        data = self._get_data()

        qualities = []
        for n_fold in range(n_folds):
            test_indices = folds[n_fold]
            train_indices = npdelete(indices, test_indices)
            self.algo.learn(samples=train_indices)
            test_data = data[test_indices]
            predictions = self.algo.predict(test_data)
            quality = self._compute_measure(test_data, predictions, multioutput)
            qualities.append(quality)

        quality = sum(qualities) / len(qualities)

        return quality

    def evaluate_bootstrap(
        self,
        n_replicates=100,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        if isinstance(samples, list):
            n_samples = len(samples)
        else:
            n_samples = samples.size
        indices = arange(n_samples)

        data = self._get_data()

        qualities = []
        for _ in range(n_replicates):
            train_indices = unique(choice(n_samples, n_samples))
            test_indices = npdelete(indices, train_indices)
            self.algo.learn(samples=[samples[index] for index in train_indices])
            test_data = data[[samples[index] for index in test_indices]]
            predictions = self.algo.predict(test_data)

            quality = self._compute_measure(test_data, predictions, multioutput)
            qualities.append(quality)

        quality = sum(qualities) / len(qualities)

        return quality

    def _compute_measure(
        self,
        data,  # type: ndarray
        labels,  # type: ndarray
        multioutput=True,  # type: bool
    ):  # type: (...) -> NoReturn
        """Compute the quality measure.

        Args:
            data: The reference data.
            labels: The predicted labels.
            multioutput: If True, return the quality measure for each
                output component. Otherwise, average these measures.

        Returns:
            The value of the quality measure.
        """
        raise NotImplementedError

    def _get_data(self):  # type: (...) -> Dict[str,ndarray]
        """Get data.

        Returns:
            The learning data indexed by the names of the variables.
        """
        names = self.algo.var_names
        data = self.algo.learning_set.get_data_by_names(names, False)
        return data



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Chains of disciplines: sequential and parallel execution processes
******************************************************************
"""
from __future__ import division, unicode_literals

import logging
from copy import deepcopy

from numpy import dot, ndarray, zeros

from gemseo.core.coupling_structure import DependencyGraph
from gemseo.core.discipline import MDODiscipline
from gemseo.core.execution_sequence import ExecutionSequenceFactory
from gemseo.core.jacobian_assembly import JacobianAssembly
from gemseo.core.parallel_execution import (
    DiscParallelExecution,
    DiscParallelLinearization,
)

LOGGER = logging.getLogger(__name__)


class MDOChain(MDODiscipline):
    """Chain of processes that is based on a predefined order of execution."""

    AVAILABLE_MODES = [
        JacobianAssembly.DIRECT_MODE,
        JacobianAssembly.REVERSE_MODE,
        JacobianAssembly.AUTO_MODE,
    ]

    def __init__(
        self, disciplines, name=None, grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE
    ):
        """Constructor of the chain.

        :param disciplines: the disciplines list
        :param name: the name of the discipline
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        """
        super(MDOChain, self).__init__(name, grammar_type=grammar_type)
        self.disciplines = disciplines
        self.initialize_grammars()
        self.default_inputs = {}
        self._update_default_inputs()

    def set_disciplines_statuses(self, status):
        """Sets the sub disciplines statuses.

        :param status: the status
        """
        for discipline in self.disciplines:
            discipline.status = status
            discipline.set_disciplines_statuses(status)

    def initialize_grammars(self):
        """Defines all inputs and outputs of the chain."""
        self.input_grammar.clear()
        self.output_grammar.clear()
        for discipline in self.disciplines:
            self.input_grammar.update_from_if_not_in(
                discipline.input_grammar, self.output_grammar
            )
            self.output_grammar.update_from(discipline.output_grammar)

    def _update_default_inputs(self):
        """Computes the default inputs from the disciplines default inputs."""
        self_inputs = self.get_input_data_names()
        for disc in self.disciplines:
            for key, value in disc.default_inputs.items():
                if key in self_inputs:
                    self.default_inputs[key] = value

    def _run(self):
        """Run discipline."""
        for discipline in self.disciplines:
            outs = discipline.execute(self.local_data)
            self.local_data.update(outs)

    def reverse_chain_rule(self, chain_outputs, discipline):
        """Chains derivatives of self, with a new discipline in the chain in reverse
        mode.

        Performs chain ruling:
        (notation: D is total derivative, d is partial derivative)

        D out    d out      dinpt_1    d output      dinpt_2
        -----  = -------- . ------- + -------- . --------
        D new_in  d inpt_1  d new_in   d inpt_2   d new_in


        D out    d out        d out      dinpt_2
        -----  = -------- + -------- . --------
        D z      d z         d inpt_2     d z


        D out    d out      [dinpt_1   d out      d inpt_1    dinpt_2 ]
        -----  = -------- . [------- + -------- . --------  . --------]
        D z      d inpt_1   [d z       d inpt_1   d inpt_2     d z    ]

        :param discipline: new discipline to compose in the chain
        :param chain_outputs: the chain_outputs to linearize
        """
        # TODO : only linearize wrt needed inputs/inputs
        # use coupling_structure graph path for that
        last_cached = discipline.cache.get_last_cached_inputs()
        discipline.linearize(last_cached, force_no_exec=True, force_all=True)

        for output in chain_outputs:
            if output in self.jac:
                # This output has already been taken from previous disciplines
                # Derivatives must be composed using the chain rule

                # Make a copy of the keys because the dict is changed in the
                # loop
                existing_inputs = self.jac[output].keys()
                common_inputs = set(existing_inputs) & set(discipline.jac)
                for input_name in common_inputs:
                    # Store reference to the current Jacobian
                    curr_j = self.jac[output][input_name]
                    for new_in, new_jac in discipline.jac[input_name].items():
                        # Chain rule the derivatives
                        # TODO: sum BEFORE dot
                        loc_dot = dot(curr_j, new_jac)
                        # when input_name==new_in, we are in the case of an
                        # input being also an output
                        # in this case we must only compose the derivatives
                        if new_in in self.jac[output] and input_name != new_in:
                            # The output is already linearized wrt this
                            # input_name. We are in the case:
                            # d o     d o    d o     di_2
                            # ----  = ---- + ----- . -----
                            # d z     d z    d i_2    d z
                            self.jac[output][new_in] += loc_dot
                        else:
                            # The output is not yet linearized wrt this
                            # input_name.  We are in the case:
                            #  d o      d o     di_1   d o     di_2
                            # -----  = ------ . ---- + ----  . ----
                            #  d x      d i_1   d x    d i_2    d x
                            self.jac[output][new_in] = loc_dot

            elif output in discipline.jac:
                # Output of the chain not yet filled in jac,
                # Take the jacobian dict of the current discipline to
                # Initialize. Make a copy !
                self.jac[output] = MDOChain.copy_jacs(discipline.jac[output])

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Actual computation of the jacobians.

        :param inputs: linearization should be performed with respect
            to inputs list.
        :param outputs: linearization should be performed on
            chain_outputs list.
        """

        # Initializes self jac with copy of last discipline (reverse mode)
        last_discipline = self.disciplines[-1]
        # TODO : only linearize wrt needed inputs/inputs
        # use coupling_structure graph path for that
        last_cached = last_discipline.cache.get_last_cached_inputs()
        last_discipline.linearize(last_cached, force_no_exec=True, force_all=True)
        self.jac = self.copy_jacs(last_discipline.jac)

        # reverse mode of remaining disciplines
        remaining_disciplines = self.disciplines[:-1]
        for discipline in remaining_disciplines[::-1]:
            self.reverse_chain_rule(outputs, discipline)

        # Remove differentiations that should not be there,
        # because inputs are not inputs of the chain
        for in_dict in self.jac.values():
            # Copy keys because the dict in changed in the loop
            input_keys_cp = list(in_dict.keys())
            for input_name in input_keys_cp:
                if input_name not in inputs:
                    del in_dict[input_name]

        # Add differentiations that should be there,
        # because inputs inputs of the chain but not
        # of all disciplines
        for out_name, jac_loc in self.jac.items():
            n_outs = len(self.get_outputs_by_name(out_name))
            for input_name in inputs:
                if input_name not in jac_loc:
                    n_inpts = len(self.get_inputs_by_name(input_name))
                    jac_loc[input_name] = zeros((n_outs, n_inpts))

    @staticmethod
    def copy_jacs(jac_dict):
        """Hard copies Jacobian dict.

        :param jac_dict: dict of dict of ndarrays, or dict of ndarrays
        :returns: deepcopy of the input
        """
        out_d = {}

        for outpt, sub_jac in jac_dict.items():
            if isinstance(sub_jac, dict):
                loc_jac_cp = {}
                out_d[outpt] = loc_jac_cp
                for inpt, jac in sub_jac.items():
                    loc_jac_cp[inpt] = jac.copy()
            elif isinstance(sub_jac, ndarray):
                out_d[outpt] = sub_jac.copy()
        return out_d

    def reset_statuses_for_run(self):
        """Sets all the statuses to PENDING."""
        super(MDOChain, self).reset_statuses_for_run()
        for discipline in self.disciplines:
            discipline.reset_statuses_for_run()

    def get_expected_workflow(self):
        """Returns the expected execution sequence, used for xdsm representation.

        See MDOFormulation.get_expected_workflow.
        """
        sequence = ExecutionSequenceFactory.serial()
        for disc in self.disciplines:
            sequence.extend(disc.get_expected_workflow())
        return sequence

    def get_expected_dataflow(self):
        """Returns the expected data exchange sequence, used for xdsm representation See
        MDOFormulation.get_expected_dataflow."""
        all_disc = list(set(self.disciplines))
        graph = DependencyGraph(all_disc)
        res = graph.get_disciplines_couplings()

        # Add discipline inner couplings (ex. MDA case)
        for disc in all_disc:
            res.extend(disc.get_expected_dataflow())

        return res

    def _set_cache_tol(self, cache_tol):
        """Sets to the cache input tolerance To be overloaded by subclasses.

        :param cache_tol: float, cache tolerance
        """
        super(MDOChain, self)._set_cache_tol(cache_tol)
        for disc in self.disciplines:
            disc.cache_tol = cache_tol or 0.0


class MDOParallelChain(MDODiscipline):
    """Chain of processes that executes disciplines in parallel."""

    def __init__(
        self,
        disciplines,
        name=None,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        use_threading=True,
        n_processes=None,
    ):
        """Constructor of the chain.

        :param disciplines: the disciplines list
        :param name: the name of the discipline
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :param use_threading: if True, use Threads instead of processes
            to parallelize the execution
        :param n_processes: maximum number of processors on which to run,
            by default the number of disciplines
        """
        super(MDOParallelChain, self).__init__(name, grammar_type=grammar_type)
        self.disciplines = disciplines
        self.initialize_grammars()
        self.default_inputs = {}
        self._update_default_inputs()
        if n_processes is None:
            n_processes = len(self.disciplines)
        dpe = DiscParallelExecution(
            self.disciplines, n_processes, use_threading=use_threading
        )
        self.parallel_execution = dpe
        dpl = DiscParallelLinearization(
            self.disciplines, n_processes, use_threading=use_threading
        )
        self.parallel_lin = dpl

    def initialize_grammars(self):
        """Defines all inputs and outputs of the chain."""
        self.input_grammar.clear()
        self.output_grammar.clear()
        for discipline in self.disciplines:
            self.input_grammar.update_from(discipline.input_grammar)
            self.output_grammar.update_from(discipline.output_grammar)

    def _update_default_inputs(self):
        """Computes the default inputs from the disciplines default inputs."""
        self_inputs = self.get_input_data_names()
        for disc in self.disciplines:
            for key, value in disc.default_inputs.items():
                if key in self_inputs:
                    self.default_inputs[key] = value

    def _get_inputs_list(self):
        """Returns a list of inputs dict for parallel execution."""
        n_disc = len(self.disciplines)
        # Avoid overlaps with dicts in // by doing a deepcopy
        # The outputs of a discipline may be a coupling, and shall therefore
        # not be passed as input of another since the execution are assumed
        # to be independent here
        all_inpts = [deepcopy(self.local_data) for _ in range(n_disc)]
        return all_inpts

    def _run(self):
        """Run discipline."""
        all_inpts = self._get_inputs_list()
        self.parallel_execution.execute(all_inpts)

        # Update data according to input order of priority
        for discipline in self.disciplines:
            out_set = discipline.get_output_data_names()
            disc_data = discipline.local_data
            out_dict = {out_k: disc_data[out_k] for out_k in out_set}
            self.local_data.update(out_dict)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Actual computation of the jacobians.

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization
            should be performed wrt all inputs  (Default value = None)
        :param outputs: linearization should be performed on
            chain_outputs list.
            If None, linearization should be
            performed on all chain_outputs (Default value = None)
        """
        self._set_disciplines_diff_outputs(outputs)
        self._set_disciplines_diff_inputs(inputs)
        all_inpts = self._get_inputs_list()
        jacobians = self.parallel_lin.execute(all_inpts)
        self.jac = {}
        # Update jacobians according to input order of priority
        for disc_jacobian in jacobians:
            for out_k, jac_loc in disc_jacobian.items():
                self_jac_o = self.jac.get(out_k)
                if self_jac_o is None:
                    self_jac_o = {}
                    self.jac[out_k] = self_jac_o
                self_jac_o.update(jac_loc)
        self._init_jacobian(inputs, outputs, with_zeros=True, fill_missing_keys=True)

    def add_differentiated_inputs(self, inputs=None):
        MDODiscipline.add_differentiated_inputs(self, inputs)
        self._set_disciplines_diff_inputs(inputs)

    def _set_disciplines_diff_inputs(self, inputs):
        """Adds inputs to the right sub discipline's differentiated inputs.

        :param inputs: the inputs list
        """
        diff_inpts = set(inputs)
        for disc in self.disciplines:
            inpt_set = set(disc.get_input_data_names()) & diff_inpts
            if inpt_set:
                disc.add_differentiated_inputs(list(inpt_set))

    def add_differentiated_outputs(self, outputs=None):
        MDODiscipline.add_differentiated_outputs(self, outputs)
        self._set_disciplines_diff_outputs(outputs)

    def _set_disciplines_diff_outputs(self, outputs):
        """Adds outputs to the right sub discipline's differentiated outputs.

        :param outputs: the outputs list
        """
        diff_outpts = set(outputs)
        for disc in self.disciplines:
            outpt_set = set(disc.get_output_data_names()) & diff_outpts
            if outpt_set:
                disc.add_differentiated_outputs(list(outpt_set))

    def reset_statuses_for_run(self):
        """Sets all the statuses to PENDING."""
        super(MDOParallelChain, self).reset_statuses_for_run()
        for discipline in self.disciplines:
            discipline.reset_statuses_for_run()

    def get_expected_workflow(self):
        """Returns the expected execution sequence, used for xdsm representation.

        See MDOFormulation.get_expected_workflow.
        """
        sequence = ExecutionSequenceFactory.parallel()
        for disc in self.disciplines:
            sequence.extend(disc.get_expected_workflow())
        return sequence

    def get_expected_dataflow(self):
        """Returns the expected data exchange sequence, used for xdsm representation See
        MDOFormulation.get_expected_dataflow."""
        return []

    def _set_cache_tol(self, cache_tol):
        """Sets to the cache input tolerance To be overloaded by subclasses.

        :param cache_tol: float, cache tolerance
        """
        super(MDOParallelChain, self)._set_cache_tol(cache_tol)
        for disc in self.disciplines:
            disc.cache_tol = cache_tol or 0.0


class MDOAdditiveChain(MDOParallelChain):
    """Chain of processes that executes disciplines in parallel and sums specified
    outputs across disciplines."""

    def __init__(
        self,
        disciplines,
        outputs_to_sum,
        name=None,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        use_threading=True,
        n_processes=None,
    ):
        """Constructor.

        :param disciplines: the disciplines list
        :type disciplines: list(MDODiscipline)
        :param outputs_to_sum: names list of the outputs to sum
        :type outputs_to_sum: list(str)
        :param name: name of the discipline
        :type name: str, optional
        :param grammar_type: the type of grammar to use for IO declaration
                            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :type grammar_type: str, optional
        :param use_threading: if True, use Threads instead of processes
            to parallelize the execution
        :type use_threading: bool, optional
        :param n_processes: maximum number of processors on which to run,
            by default the number of disciplines
        :type n_processes: int, optional
        """
        super(MDOAdditiveChain, self).__init__(
            disciplines, name, grammar_type, use_threading, n_processes
        )
        self._outputs_to_sum = outputs_to_sum

    def _run(self):
        """Runs the disciplines and computes the sum."""
        # Run the disciplines in parallel
        MDOParallelChain._run(self)

        # Sum the required outputs across disciplines
        for out_name in self._outputs_to_sum:
            terms = [
                disc.local_data[out_name]
                for disc in self.disciplines
                if out_name in disc.local_data
            ]
            sum_value = sum(terms) if terms else None
            self.local_data[out_name] = sum_value

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Actual computation of the Jacobians.

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization
            should be performed wrt all inputs  (Default value = None)
        :param outputs: linearization should be performed on
            chain_outputs list.
            If None, linearization should be
            performed on all chain_outputs (Default value = None)
        """
        # Differentiate the disciplines in parallel
        MDOParallelChain._compute_jacobian(self, inputs, outputs)

        # Sum the Jacobians of the required outputs across disciplines
        for out_name in self._outputs_to_sum:
            self.jac[out_name] = dict()
            for in_name in inputs:
                terms = [
                    disc.jac[out_name][in_name]
                    for disc in self.disciplines
                    if in_name in disc.jac[out_name]
                ]

                assert terms
                sum_value = sum(terms)
                self.jac[out_name][in_name] = sum_value



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable problem - Models
*************************
"""
from __future__ import division, unicode_literals

import logging

from numpy import array, atleast_2d, diag, dot, eye
from numpy import mean as npmean
from numpy import ndarray
from numpy import sum as npsum
from past.utils import old_div

from .variables import (
    OBJECTIVE_NAME,
    X_SHARED_NAME,
    get_constraint_name,
    get_coupling_name,
    get_u_local_name,
    get_x_local_name,
)

LOGGER = logging.getLogger(__name__)


class TMMainModel(object):

    r"""The main discipline from the scalable problem introduced by Tedford
    and Martins (2010) takes the  local design parameters
    :math:`x_1,x_2,\ldots,x_N` and the global design parameters :math:`z`
    as inputs, as well as the coupling variables :math:`y_1,y_2,\ldots,y_N`
    and returns the objective function value :math:`f(x,y(x,y))` to minimize as
    well as the inequality constraints ones
    :math:`c_1(y_1),c_2(y_2),\ldots,c_N(y_N)` which are expressed as:

    .. math::

       f(z,y) = |z|_2^2 + \sum_{i=1}^N |y_i|_2^2


    and:

    .. math::

       c_i(y_i) = 1- C_i^{-T}Iy_i

    """

    def __init__(self, c_constraint, default_inputs):
        """Constructor.

        :param list(ndarray) c_constraint: constraint coefficients
        :param dict(ndarray) default_inputs: default inputs
        """
        self.name = "MainModel"
        self.default_inputs = default_inputs
        self.inputs_sizes = {name: len(val) for name, val in default_inputs.items()}
        self.inputs_names = sorted(self.inputs_sizes.keys())
        self.outputs_sizes = {
            get_constraint_name(index): len(value)
            for index, value in enumerate(c_constraint)
        }
        self.outputs_sizes[OBJECTIVE_NAME] = 1
        self.outputs_names = sorted(self.outputs_sizes.keys())
        self.coefficients = c_constraint
        self.n_submodels = len(c_constraint)

    def __call__(self, x_shared=None, coupling=None, jacobian=False):
        """Compute the discipline output or derivatives for new values of shared design
        parameters and coupling variables.

        :param ndarray x_shared: shared design parameters.
        :param dict(ndarray) coupling: list of coupling variables
            (one element per sub-discipline).
        :param bool jacobian: if True, return the jacobian of the discipline output.
            Otherwise, return the discipline output.
        """
        if x_shared is None:
            x_shared = self.default_inputs[X_SHARED_NAME]
        if coupling is None:
            names = set(self.inputs_names) - {X_SHARED_NAME}
            coupling = {name: self.default_inputs[name] for name in names}
        if jacobian:
            result = self._compute_jacobian(x_shared, coupling)
        else:
            result = self._compute_output(x_shared, coupling)
        return result

    def _compute_output(self, x_shared, coupling):
        """Compute the discipline output for new values of shared design parameters and
        coupling variables.

        :param ndarray x_shared: shared design parameters.
        :param dict(ndarray) coupling: list of coupling variables
            (one element per sub-discipline).
        """
        obj = npmean(x_shared ** 2)
        obj += npmean(
            [
                npmean(coupling[get_coupling_name(index)] ** 2)
                for index in range(self.n_submodels)
            ]
        )
        output = {OBJECTIVE_NAME: array([obj])}
        for index in range(self.n_submodels):
            constraint = get_constraint_name(index)
            coupling_name = get_coupling_name(index)
            tmp = old_div(coupling[coupling_name], self.coefficients[index])
            output[constraint] = 1 - tmp
        return output

    def _compute_jacobian(self, x_shared, coupling):
        """Computes the discipline jacobian.

        :param ndarray x_shared: shared design parameters.
        :param dict(ndarray) coupling: list of coupling variables
            (one element per sub-discipline).
        """
        tmp = old_div(2 * x_shared, len(x_shared))
        jacobian = {OBJECTIVE_NAME: {}}
        jacobian[OBJECTIVE_NAME][X_SHARED_NAME] = atleast_2d(tmp)
        for index in range(self.n_submodels):
            constraint = get_constraint_name(index)
            coupling_name = get_coupling_name(index)
            jacobian[constraint] = {}
            tmp = old_div(2 * coupling[coupling_name], len(coupling[coupling_name]))
            tmp /= self.n_submodels
            jacobian[OBJECTIVE_NAME][coupling_name] = atleast_2d(tmp)
            tmp = diag(old_div(-1.0, self.coefficients[index]))
            jacobian[constraint][coupling_name] = atleast_2d(tmp)
        return jacobian


class TMSubModel(object):

    r"""A sub-discipline from the scalable problem introduced by
    Tedford and Martins (2010) takes local design parameters :math:`x_i`
    and shared design parameters :math:`z` in input as well as coupling
    variables :math:`\left(y_i\right)_{1\leq j \leq N\atop j\neq i}`
    from :math:`N-1` elementary disciplines,
    and returns the coupling variables:

    .. math::

        y_i =\frac{\tilde{y}_i+C_{z,i}.1+C_{x_i}.1}{\sum_{j=1 \atop j
        \neq i}^NC_{y_j,i}.1+C_{z,i}.1+C_{x_i}.1} \in [0,1]^{n_{y_i}}

    where:

    .. math::

        \tilde{y}_i = - C_{z,i}.z - C_{x_i}.x_i +
        \sum_{j=1 \atop j \neq i}^N C_{y_j,i}.y_j
    """

    def __init__(self, index, c_shared, c_local, c_coupling, default_inputs):
        """Constructor.

        :param int index: discipline index for naming.
        :param ndarray c_shared: weights for the shared design parameters.
        :param ndarray c_local: weights for the local design parameters.
        :param dict(ndarray) c_coupling: weights for the coupling parameters.
        :param dict(ndarray) default_inputs: default inputs
        """
        self.name = "SubModel_{}".format(index)
        self.index = index
        self.c_shared = c_shared
        self.c_local = c_local
        self.c_coupling = c_coupling
        self.default_inputs = default_inputs
        self._check_consistency()
        self.inputs_sizes = {name: len(val) for name, val in default_inputs.items()}
        self.inputs_names = sorted(self.inputs_sizes.keys())
        output = get_coupling_name(index)
        self.outputs_sizes = {output: len(c_local)}
        self.outputs_names = sorted(self.outputs_sizes.keys())

    def _check_consistency(self):
        """Check consistency of model and default inputs."""
        assert isinstance(self.c_shared, ndarray)
        assert len(self.c_shared.shape) == 2
        assert self.c_shared.shape[1] == len(self.default_inputs[X_SHARED_NAME])
        nout = self.c_shared.shape[0]
        assert isinstance(self.c_local, ndarray)
        assert len(self.c_local.shape) == 2
        assert self.c_local.shape[0] == nout
        xlocal = get_x_local_name(self.index)
        assert self.c_local.shape[1] == len(self.default_inputs[xlocal])
        assert isinstance(self.c_coupling, dict)
        for key, val in self.c_coupling.items():
            assert isinstance(val, ndarray)
            assert len(val.shape) == 2
            assert val.shape[0] == nout
            assert val.shape[1] == len(self.default_inputs[key])

    def __call__(
        self, x_shared=None, x_local=None, coupling=None, noise=None, jacobian=False
    ):
        """Compute the discipline output or derivatives for new values of shared design
        parameters and coupling variables.

        :param ndarray x_shared: shared design parameters.
        :param ndarray x_local: local design parameters.
        :param dict(ndarray) coupling: list of coupling variables
            (one element per sub-discipline).
        :param ndarray noise: random noise applied to the vectorial output.
        :param bool jacobian: if True, return the jacobian of the discipline output.
            Otherwise, return the discipline output.
        """
        if x_shared is None:
            x_shared = self.default_inputs[X_SHARED_NAME]
        if x_local is None:
            x_local = self.default_inputs[get_x_local_name(self.index)]
        if coupling is None:
            coupling = {cpl: self.default_inputs[cpl] for cpl in self.c_coupling}
        if jacobian:
            result = self._compute_jacobian()
        else:
            result = self._compute_output(x_shared, x_local, coupling, noise)
        return result

    def _compute_output(self, x_shared, x_local, coupling, noise):
        """Compute the discipline output for new values of shared design parameters and
        coupling variables.

        :param ndarray x_shared: shared design parameters.
        :param ndarray x_local: local design parameters.
        :param dict(ndarray) coupling: list of coupling variables
            (one element per sub-discipline).
        :param ndarray noise: random noise applied to the vectorial output.
        """
        output = -dot(self.c_shared, x_shared.reshape((-1, 1)))
        output -= dot(self.c_local, x_local.reshape((-1, 1)))
        cpl_sum = 0
        for name, coeff in self.c_coupling.items():
            output += dot(coeff, coupling[name].reshape((-1, 1)))
            cpl_sum += npsum(coeff, 1)
        output += npsum(self.c_shared, 1).reshape((-1, 1))
        output += npsum(self.c_local, 1).reshape((-1, 1))
        norm = cpl_sum.reshape((-1, 1))
        norm += npsum(self.c_shared, 1).reshape((-1, 1))
        norm += npsum(self.c_local, 1).reshape((-1, 1))
        output /= norm
        output = output.flatten()
        if noise is not None:
            output += noise
        output = {get_coupling_name(self.index): output}
        return output

    def _compute_jacobian(self):
        """Computes the discipline jacobian."""
        coupling_name = get_coupling_name(self.index)
        x_local_name = get_x_local_name(self.index)
        u_local_name = get_u_local_name(self.index)
        cpl_sum = 0
        for coeff in self.c_coupling.values():
            cpl_sum += npsum(coeff, 1)
        norm = cpl_sum.reshape((-1, 1))
        norm += npsum(self.c_shared, 1).reshape((-1, 1))
        norm += npsum(self.c_local, 1).reshape((-1, 1))
        der = old_div(-self.c_local, norm)
        jacobian = {coupling_name: {}}
        jacobian[coupling_name][x_local_name] = der
        jacobian[coupling_name][u_local_name] = eye(der.shape[0])
        der = old_div(-self.c_shared, norm)
        jacobian[coupling_name][X_SHARED_NAME] = der
        for cpl_name, cpl_value in self.c_coupling.items():
            jacobian[coupling_name][cpl_name] = old_div(cpl_value, norm)
        return jacobian



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Jean-Christophe Giret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Custom Enumerations.

The `.BaseEnum` class enables the specification of options which were previously defined
using class attributes. It enables backward compatibility, as the user can either
provide the Enum member or its name as a string. The conversion is then made implicitly.
"""

from __future__ import absolute_import

from enum import Enum, EnumMeta
from typing import Union

from six import with_metaclass


class MetaEnum(EnumMeta):
    """An Enum meta-class to subclass the `in` behavior."""

    def __contains__(cls, item):  # noqa: N805  # type: (...) -> bool
        if isinstance(item, cls):
            return super(MetaEnum, cls).__contains__(item)
        elif isinstance(item, str):
            return item in cls.__members__.keys()
        return False


class BaseEnum(with_metaclass(MetaEnum, Enum)):
    """A base Enum class that can be compared to strings."""

    @classmethod
    def get_member_from_name(
        cls,
        value,  # type: Union[str, BaseEnum]
    ):  # type: (...) -> BaseEnum
        """Return an Enum member from a name or a member.

        This class method returns an Enum member
        either from the name of the member or the member itself.
        It enables backward compatibility with the use of class attributes
        for options.

        Args:
            value: An Enum member name or an Enum member.

        Raises:
            TypeError: If the Enum member is not from the same Enum class.
        """
        if isinstance(value, str):
            return cls.__members__[value]
        elif isinstance(value, cls):
            return value

        raise TypeError(
            "The type of value is {} but {} or str are expected.".format(
                type(value), str(cls)
            )
        )

    def __eq__(
        self,
        other,  # type: Union[BaseEnum, str]
    ):  # type: (...) -> bool
        if isinstance(other, self.__class__):
            return other.value == self.value
        elif isinstance(other, str):
            return other == self.name
        return False

    def __str__(self):  # type: (...) -> str
        return self.name


class CamelCaseEnum(Enum):
    """Enum that are represented as the camel case of the key name."""

    def __str__(self):  # type: (...) -> str
        return self.name.title().replace("_", "")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#         Francois Gallard : refactoring for v1, May 2016
"""scipy.optimize optimization library wrapper."""
from __future__ import division, unicode_literals

import logging

from numpy import isfinite, real
from scipy import optimize

from gemseo.algos.opt.opt_lib import OptimizationLibrary

LOGGER = logging.getLogger(__name__)


class ScipyOpt(OptimizationLibrary):
    """Scipy optimization library interface.

    See OptimizationLibrary.
    """

    LIB_COMPUTE_GRAD = True

    OPTIONS_MAP = {
        # Available only in the doc !
        # OptimizationLibrary.LS_STEP_NB_MAX: "maxls",
        OptimizationLibrary.LS_STEP_SIZE_MAX: "stepmx",
        OptimizationLibrary.MAX_FUN_EVAL: "maxfun",
        OptimizationLibrary.PG_TOL: "gtol",
    }

    def __init__(self):
        """Constructor.

        Generate the library dict, contains the list
        of algorithms with their characteristics:

        - does it require gradient
        - does it handle equality constraints
        - does it handle inequality constraints
        """
        super(ScipyOpt, self).__init__()
        doc = "https://docs.scipy.org/doc/scipy/reference/"
        self.lib_dict = {
            "SLSQP": {
                self.INTERNAL_NAME: "SLSQP",
                self.REQUIRE_GRAD: True,
                self.POSITIVE_CONSTRAINTS: True,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "Sequential Least-Squares Quadratic "
                "Programming (SLSQP) implemented in "
                "the SciPy library",
                self.WEBSITE: doc + "optimize.minimize-slsqp.html",
            },
            "L-BFGS-B": {
                self.INTERNAL_NAME: "L-BFGS-B",
                self.REQUIRE_GRAD: True,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "Limited-memory BFGS algorithm "
                "implemented in SciPy library",
                self.WEBSITE: doc + "generated/scipy.optimize.fmin_l_bfgs_b.html",
            },
            "TNC": {
                self.INTERNAL_NAME: "TNC",
                self.REQUIRE_GRAD: True,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "Truncated Newton (TNC) algorithm "
                "implemented in SciPy library",
                self.WEBSITE: doc + "optimize.minimize-tnc.html",
            },
        }

    def _get_options(  # pylint: disable=W0221
        self,
        max_iter=999,
        ftol_rel=1e-9,
        ftol_abs=1e-9,
        xtol_rel=1e-9,
        xtol_abs=1e-9,
        max_ls_step_size=0.0,
        max_ls_step_nb=20,
        max_fun_eval=999,
        max_time=0,
        pg_tol=1e-5,
        disp=0,
        max_c_git=-1,
        eta=-1.0,
        factr=1e7,
        maxcor=20,
        normalize_design_space=True,
        eq_tolerance=1e-2,
        ineq_tolerance=1e-4,
        stepmx=0.0,
        minfev=0.0,
        scale=None,
        rescale=None,
        offset=None,
        **kwargs
    ):
        r"""Sets the options default values

        To get the best and up to date information about algorithms options,
        go to scipy.optimize documentation:
        https://docs.scipy.org/doc/scipy/reference/tutorial/optimize.html

        :param max_iter: maximum number of iterations, ie unique calls to f(x)
        :type max_iter: int
        :param ftol_rel: stop criteria, relative tolerance on the
               objective function,
               if abs(f(xk)-f(xk+1))/abs(f(xk))<= ftol_rel: stop
               (Default value = 1e-9)
        :type ftol_rel: float
        :param ftol_abs: stop criteria, absolute tolerance on the objective
               function, if abs(f(xk)-f(xk+1))<= ftol_rel: stop
               (Default value = 1e-9)
        :type ftol_abs: float
        :param xtol_rel: stop criteria, relative tolerance on the
               design variables,
               if norm(xk-xk+1)/norm(xk)<= xtol_rel: stop
               (Default value = 1e-9)
        :type xtol_rel: float
        :param xtol_abs: stop criteria, absolute tolerance on the
               design variables,
               if norm(xk-xk+1)<= xtol_abs: stop
               (Default value = 1e-9)
        :type xtol_abs: float
        :param max_ls_step_size: Maximum step for the line search
               (Default value = 0.)
        :type max_ls_step_size: float
        :param max_ls_step_nb: Maximum number of line search steps
               per iteration. (Default value = 20)
        :type max_ls_step_nb: int
        :param max_fun_eval: internal stop criteria on the
               number of algorithm outer iterations (Default value = 999)
        :type max_ls_step_size: int
        :param max_time: maximum runtime in seconds,
            disabled if 0 (Default value = 0)
        :type max_time: float
        :param pg_tol: stop criteria on the projected gradient norm
               (Default value = 1e-5)
        :type pg_tol: float
        :param disp: display information, (Default value = 0)
        :type disp: int
        :param max_c_git: Maximum Conjugate Gradient internal solver
               iterations (Default value = -1)
        :type max_c_git: int
        :param eta: severity of the linesearch, specific to
               TNC algorithm (Default value = -1.)
        :type eta: int
        :param factr: stop criteria on the projected gradient norm,
               stop if max_i (grad_i)<eps_mach \* factr, where eps_mach is the
               machine precision ( Default value = 1e7)
        :type factr: float
        :param maxcor: maximum BFGS updates (Default value = 20)
        :type maxcor: int
        :param normalize_design_space: If True, scales variables in [0, 1]
        :type normalize_design_space: bool
        :param eq_tolerance: equality tolerance
        :type eq_tolerance: float
        :param ineq_tolerance: inequality tolerance
        :type ineq_tolerance: float
        :param stepmx: Maximum step for the line search.
        :type stepmx: float
        :param minfev: Minimum function value estimate
        :type minfev: float
        :param scale: Scaling factors to apply to each variable
        :type scale: array
        :param rescale: Scaling factor (in log10) used to trigger f value
            rescaling
        :type rescale: float
        :param offset: Value to subtract from each variable.
        :type offset: float
        :param kwargs: other algorithms options
        :type kwargs: kwargs
        """
        nds = normalize_design_space
        popts = self._process_options(
            max_iter=max_iter,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            xtol_abs=xtol_abs,
            max_time=max_time,
            max_ls_step_size=max_ls_step_size,
            max_ls_step_nb=max_ls_step_nb,
            max_fun_eval=max_fun_eval,
            pg_tol=pg_tol,
            disp=disp,
            max_c_git=max_c_git,
            eta=eta,
            factr=factr,
            maxcor=maxcor,
            normalize_design_space=nds,
            ineq_tolerance=ineq_tolerance,
            eq_tolerance=eq_tolerance,
            stepmx=stepmx,
            minfev=minfev,
            scale=scale,
            rescale=rescale,
            offset=offset,
            **kwargs
        )
        return popts

    def _run(self, **options):
        """Runs the algorithm, to be overloaded by subclasses.

        :param options: the options dict for the algorithm
        """
        # remove normalization from options for algo
        normalize_ds = options.pop(self.NORMALIZE_DESIGN_SPACE_OPTION, True)
        # Get the normalized bounds:
        x_0, l_b, u_b = self.get_x0_and_bounds_vects(normalize_ds)
        # Replace infinite values with None:
        l_b = [val if isfinite(val) else None for val in l_b]
        u_b = [val if isfinite(val) else None for val in u_b]
        bounds = list(zip(l_b, u_b))

        def real_part_fun(x_vect):
            """Wraps the function and returns the real part."""
            return real(self.problem.objective.func(x_vect))

        fun = real_part_fun

        constraints = self.get_right_sign_constraints()
        cstr_scipy = []
        for cstr in constraints:
            c_scipy = {"type": cstr.f_type, "fun": cstr.func, "jac": cstr.jac}
            cstr_scipy.append(c_scipy)
        jac = self.problem.objective.jac

        # |g| is in charge of ensuring max iterations, and
        # xtol, ftol, since it may
        # have a different definition of iterations, such as for SLSQP
        # for instance which counts duplicate calls to x as a new iteration
        options["maxiter"] = 10000000

        # Deactivate scipy stop criteria to use |g|' ones
        options["ftol"] = 0.0
        options["xtol"] = 0.0
        options.pop(self.F_TOL_ABS)
        options.pop(self.X_TOL_ABS)
        options.pop(self.F_TOL_REL)
        options.pop(self.X_TOL_REL)
        options.pop(self.MAX_TIME)
        options.pop(self.MAX_ITER)
        if self.algo_name != "TNC":
            options.pop("xtol")
        opt_result = optimize.minimize(
            fun=fun,
            x0=x_0,
            method=self.internal_algo_name,
            jac=jac,
            bounds=bounds,
            constraints=cstr_scipy,
            tol=None,
            options=options,
        )

        return self.get_optimum_from_database(opt_result.message, opt_result.status)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#    initial documentation
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
# from Tedford2010
# Propane combustion, p12
r"""
The propane combustion MDO problem
**********************************

The Propane MDO problem can be found in :cite:`Padula1996`
and :cite:`TedfordMartins2006`. It represents the chemical equilibrium
reached during the combustion of propane in air.
Variables are assigned to represent each of the ten combustion products
as well as the sum of the products.

The optimization problem is as follows:


.. math::

   \begin{aligned}
   \text{minimize the objective function } &
   f_2 + f_6 + f_7 + f_9 \\
   \text{with respect to the design variables } &
   x_{1},\,x_{3},\,x_{6},\,x_{7} \\
   \text{subject to the general constraints }
   & f_2(x) \geq 0\\
   & f_6(x) \geq 0\\
   & f_7(x) \geq 0\\
   & f_9(x) \geq 0\\
   \text{subject to the bound constraints }
   & x_{1} \geq 0\\
   & x_{3} \geq 0\\
   & x_{6} \geq 0\\
   & x_{7} \geq 0\\
   \end{aligned}

where the System Discipline consists of computing the following expressions:

.. math::

   \begin{aligned}
   f_2(x) & = & 2x_1 + x_2 + x_4 + x_7 + x_8 + x_9 + 2x_{10} - R, \\
   f_6(x) & = & K_6x_2^{1/2}x_4^{1/2} - x_1^{1/2}x_6(p/x_{11})^{1/2}, \\
   f_7(x) & = & K_7x_1^{1/2}x_2^{1/2} - x_4^{1/2}x_7(p/x_{11})^{1/2}, \\
   f_9(x) & = & K_9x_1x_3^{1/2} - x_4x_9(p/x_{11})^{1/2}. \\
   \end{aligned}


Discipline 1 computes :math:`(x_{2}, x_{4})`
by satisfying the following equations:

.. math::

   \begin{aligned}
   x_1 + x_4 - 3 &=& 0,\\
   K_5x_2x_4 - x_1x_5 &=& 0.\\
   \end{aligned}

Discipline 2 computes :math:`(x_2, x_4)` such that:

.. math::

   \begin{aligned}
   K_8x_1 + x_4x_8(p/x_{11}) &=& 0,\\
   K_{10}x_{1}^{2} - x_4^2x_{10}(p/x_{11}) &=& 0.\\
   \end{aligned}

and Discipline 3 computes :math:`(x_5, x_9, x_{11})` by solving:

.. math::

   \begin{aligned}
   2x_2 + 2x_5 + x_6 + x_7 - 8&=& 0,\\
   2x_3 + x_9 - 4R &=& 0, \\
   x_{11} - \sum_{j=1}^{10} x_j &=& 0. \\
   \end{aligned}
"""
from __future__ import division, unicode_literals

from cmath import sqrt
from os.path import dirname, join

from numpy import array, complex128, ones, zeros

from gemseo.algos.design_space import DesignSpace
from gemseo.core.discipline import MDODiscipline


def get_design_space(to_complex=True):
    """Reads the design space file.

    :param to_complex: if True, current x is a complex vector
    """
    ds_read = DesignSpace.read_from_txt(
        join(dirname(__file__), "propane_design_space.txt")
    )
    if to_complex:
        ds_read.to_complex()
    return ds_read


class PropaneReaction(MDODiscipline):

    """Propane's objective and constraints discipline This discipline's outputs are the
    objective function and partial terms used in inequality constraints.

    Note: the equations have been decoupled (y_i = y_i(x_shared)). Otherwise,
    the solvers may find iterates for
    which discipline analyses are not computable.
    """

    def __init__(self):
        super(PropaneReaction, self).__init__(auto_detect_grammar_files=True)
        self.default_inputs = {
            "x_shared": ones(4, dtype=complex128),
            "y_1": ones(2, dtype=complex128),
            "y_2": ones(2, dtype=complex128),
            "y_3": ones(3, dtype=complex128),
        }
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _run(self):
        """Compute the outputs of the propane combustion model."""
        inputs = ["y_1", "y_2", "y_3", "x_shared"]
        y_1, y_2, y_3, x_shared = self.get_inputs_by_name(inputs)
        f2_list = [self.f_2(x_shared, y_1, y_2, y_3)]
        f_2 = array(f2_list, dtype=complex128)
        f_6 = array([self.f_6(x_shared, y_1, y_3)], dtype=complex128)
        f_7 = array([self.f_7(x_shared, y_1, y_3)], dtype=complex128)
        f_9 = array([self.f_9(x_shared, y_1, y_3)], dtype=complex128)
        obj = f_2 + f_7 + f_6 + f_9
        # constraints on f_2, f_6, f_7, f_9 must be nonnegative
        # in original problem don't forget to take the opposite
        self.store_local_data(f_2=-f_2, f_6=-f_6, f_7=-f_7, f_9=-f_9, obj=obj)

    @classmethod
    def f_2(cls, x_shared, y_1, y_2, y_3):
        """First term of a sum of four in the objective function. Is also a nonnegative
        constraint at system level.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :param y_1: first coupling variable
        :type y_1: ndarray
        :param y_2: second coupling variable
        :type y_2: ndarray
        :param y_3: third coupling variable
        :type y_3: ndarray
        :returns: f2(x_shared, y_1, y_2, y_3)
        :rtype: float
        """
        return (
            2.0 * x_shared[0]
            + y_1[0]
            + y_1[1]
            + x_shared[3]
            + y_2[1]
            + y_3[0]
            + 2.0 * y_3[1]
            - 10.0
        )

    @classmethod
    def f_6(cls, x_shared, y_1, y_3):
        """Second term of a sum of four in the objective function. Is also a nonnegative
        constraint at system level.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :param y_1: first coupling variable
        :type y_1: ndarray
        :param y_3: third coupling variable
        :type y_3: ndarray
        :returns: f6(x, y)
        :rtype: float
        """
        return sqrt(y_1[0] * y_1[1]) - sqrt(40.0 * x_shared[0] / y_3[2]) * x_shared[2]

    @classmethod
    def f_7(cls, x_shared, y_1, y_3):
        """Third term of a sum of four in the objective function. Is also a nonnegative
        constraint at system level.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :param y_1: first coupling variable
        :type y_1: ndarray
        :param y_3: third coupling variable
        :type y_3: ndarray
        :returns: f7(x, y)
        :rtype: float
        """
        return sqrt(x_shared[0] * y_1[0]) - sqrt(40.0 * y_1[1] / y_3[2]) * x_shared[3]

    @classmethod
    def f_9(cls, x_shared, y_1, y_3):
        """Fourth term of a sum of four in the objective function. Is also a nonnegative
        constraint at system level.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :param y_1: first coupling variable
        :type y_1: ndarray
        :param y_3: third coupling variable
        :type y_3: ndarray
        :returns: f9(x, y)
        :rtype: float
        """
        return x_shared[0] * sqrt(x_shared[1]) - y_1[1] * y_3[0] * sqrt(40.0 / y_3[2])


class PropaneComb1(MDODiscipline):

    """Propane combustion 1st set of equations This discipline is characterized by two
    governing equations."""

    def __init__(self):
        super(PropaneComb1, self).__init__(auto_detect_grammar_files=True)
        self.default_inputs = {"x_shared": ones(4, dtype=complex128)}
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _run(self):
        """Solve 2 coupling equations in functional form and compute coupling variables
        Y0 and y_1."""
        x_shared = self.get_inputs_by_name("x_shared")
        y_1_out = zeros(2, dtype=complex128)
        y_1_out[0] = self.compute_y0(x_shared)
        y_1_out[1] = self.compute_y1(x_shared)
        self.store_local_data(y_1=y_1_out)

    @classmethod
    def compute_y0(cls, x_shared):
        """Solve the first coupling equation in functional form.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :returns: coupling variable y0
        """
        return -x_shared[0] * (x_shared[2] + x_shared[3] - 8.0) / 6.0

    @classmethod
    def compute_y1(cls, x_shared):
        """Solve the second coupling equation in functional form.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :returns: coupling variable y1
        """
        return 3.0 - x_shared[0]


class PropaneComb2(MDODiscipline):
    """Propane combustion 2nd set of equations This discipline is characterized by two
    governing equations."""

    def __init__(self):
        super(PropaneComb2, self).__init__(auto_detect_grammar_files=True)
        self.default_inputs = {"x_shared": ones(4, dtype=complex128)}
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _run(self):
        """Solve 2 coupling equations in functional form and compute coupling variables
        y_2 and y_3."""
        x_shared = self.get_inputs_by_name("x_shared")
        y_2_out = zeros(2, dtype=complex128)
        y_2_out[0] = self.compute_y2(x_shared)
        y_2_out[1] = self.compute_y3(x_shared)
        self.store_local_data(y_2=y_2_out)

    @classmethod
    def compute_y2(cls, x_shared):
        """Solve the third coupling equation in functional form.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :returns: coupling variable y_2
        """
        return (x_shared[0] - 3.0) * (x_shared[2] + x_shared[3] - 8.0) / 6.0

    @classmethod
    def compute_y3(cls, x_shared):
        """Solve the fourth coupling equation in functional form.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :returns: coupling variable y_3
        """
        y_3 = -(x_shared[0] - 3.0) * x_shared[0]
        y_3 *= x_shared[2] + x_shared[3] - 2.0 * x_shared[1] + 94.0
        y_3 /= 2.0 * (400.0 * x_shared[0] ** 2.0 - 2403.0 * x_shared[0] + 3600.0)
        return y_3


class PropaneComb3(MDODiscipline):

    """This discipline is characterized by three governing equations."""

    def __init__(self):
        super(PropaneComb3, self).__init__(auto_detect_grammar_files=True)
        self.default_inputs = {"x_shared": ones(4, dtype=complex128)}
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _run(self):
        """Solve 3 coupling equations in functional form and compute coupling variables
        y_4, Y5 and Y6."""
        x_shared = self.get_inputs_by_name("x_shared")
        y_3_out = zeros(3, dtype=complex128)
        y_3_out[0] = self.compute_y4(x_shared)
        y_3_out[1] = self.compute_y5(x_shared)
        y_3_out[2] = self.compute_y6(x_shared)
        self.store_local_data(y_3=y_3_out)

    @classmethod
    def compute_y4(cls, x_shared):
        """Solve the fifth coupling equation in functional form.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :returns: coupling variable y_4
        """
        return 40.0 - 2.0 * x_shared[1]

    @classmethod
    def compute_y5(cls, x_shared):
        """Solve the sixth coupling equation in functional form.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :returns: coupling variable Y5
        """
        y_5 = x_shared[0] ** 2
        y_5 *= x_shared[2] + x_shared[3] - 2.0 * x_shared[1] + 94.0
        y_5 /= 2.0 * (400.0 * x_shared[0] ** 2 - 2403.0 * x_shared[0] + 3600.0)
        return y_5

    @classmethod
    def compute_y6(cls, x_shared):
        """Solve the seventh coupling equation in functional form.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :returns: coupling variable Y6
        """
        y_6 = 200.0 * (x_shared[0] - 3.0) ** 2
        y_6 *= x_shared[2] + x_shared[3] - 2.0 * x_shared[1] + 94.0
        y_6 /= 400.0 * x_shared[0] ** 2 - 2403.0 * x_shared[0] + 3600.0
        return y_6


# =========================================================================
# Analytical Jacobians of f and residuals #
# =========================================================================
#
# def jacobian_f_x(x_shared, Y):
#     """
#     Jacobian matrix of objective function wrt design variables
#     :param x_shared: vector of shared design variables
#     :type x_shared: ndarray
#     :param Y: vector of coupling variables
#     :type Y: ndarray
#     :returns: Jacobian matrix
#     """
#     return array([
#         2. - x_shared[2] * sqrt(10. * Y[6] / x_shared[0]) / Y[6] +
#         sqrt(x_shared[1]) + Y[0] / (2. * sqrt(x_shared[0] * Y[0])),
#         x_shared[0] / (2. * sqrt(x_shared[1])),
#         -2. * sqrt(10. * x_shared[0] / Y[6]),
#         1. - 2. * sqrt(10 * Y[1] / Y[6])
#     ])
# #
#
# def jacobian_f_y(x_shared, Y):
#     """
#     Jacobian matrix of objective function wrt coupling variables
#     :param x_shared: vector of shared design variables
#     :type x_shared: ndarray
#     :param Y: vector of coupling variables
#     :type Y: ndarray
#     :returns: Jacobian matrix
#     """
#     return array(
#         [Y[1] / (2. * sqrt(Y[0] * Y[1])) + x_shared[0] /
#          (2. * sqrt(x_shared[0] * Y[0])) + 1., -
#          x_shared[3] * sqrt(10. * Y[6] / Y[1]) / Y[6] - 2. *
#          Y[4] * sqrt(10. / Y[6]) +
#          Y[0] / (2. * sqrt(Y[0] * Y[1])) + 1., 0., 1., 1. - 2. * Y[1] *
#          sqrt(10. / Y[6]),
#          2.,
#          (Y[1] * x_shared[3] * sqrt(10. * Y[6] / Y[1]) + x_shared[0] *
# x_shared[2] *
#           sqrt(10. * Y[6] / x_shared[0])) / Y[6] ** 2 + Y[1] *
#          Y[4] * sqrt(10. / Y[6] ** 3)])
#
#
# def jacobian_residual_x(x_shared, Y):
#     """
#     Jacobian matrix of residual vector wrt design variables
#     :param x_shared: vector of shared design variables
#     :type x_shared: ndarray
#     :param Y: vector of coupling variables
#     :type Y: ndarray
#     :returns: Jacobian matrix
#     """
#     return array([
#         [1., 0., 0., 0.],
#         [-Y[2], 0., 0., 0.],
#         [0.1, 0., 0., 0.],
#         [0.2 * x_shared[0], 0., 0., 0.],
#         [0., 0., 1., 1.],
#         [0., 2., 0., 0.],
#         [-1., -1., -1., -1.]
#     ])
#
#
# def jacobian_residual_y(x_shared, Y):
#     """
#     Jacobian matrix of residual vector wrt coupling variables
#     :param x_shared: vector of shared design variables
#     :type x_shared: ndarray
#     :param Y: vector of coupling variables
#     :type Y: ndarray
#     :returns: Jacobian matrix
#     """
#     return array([
#         [0, 1., 0., 0., 0., 0., 0.],
#         [Y[1], Y[0], -Z[0], 0, 0, 0, 0],
#         [0., -40 * Y[3] / Y[6], 0., -40 * Y[1] / Y[6], 0., 0., 40 *
#          Y[1] * Y[3] / Y[6] ** 2],
#         [0., - 80 * Y[1] * Y[5] / Y[6], 0., 0., 0., -40 * Y[1] ** 2 / Y[6],
#          40 * Y[1] ** 2 * Y[5] / Y[6] ** 2],
#         [2., 0., 2., 0., 0., 0., 0.],
#         [0., 0., 0., 0., 1., 0., 0.],
#         [-1., -1., -1., -1., -1., -1., 1.]])



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The k-nearest neighbors for classification.

The k-nearest neighbor classification algorithm is an approach
to predict the output class of a new input point
by selecting the majority class among the k nearest neighbors in a training set
through voting.
The algorithm may also predict the probabilities of belonging to each class
by counting the number of occurrences of the class withing the k nearest neighbors.

Let :math:`(x_i)_{i=1,\\cdots,n_{\\text{samples}}}\\in
\\mathbb{R}^{n_{\\text{samples}}\\times n_{\\text{inputs}}}`
and :math:`(y_i)_{i=1,\\cdots,n_{\\text{samples}}}\\in
\\{1,\\cdots,n_{\\text{classes}}\\}^{n_{\\text{samples}}}`
denote the input and output training data respectively.

The procedure for predicting the class of a new input point :math:`x\\in
\\mathbb{R}^{n_{\\text{inputs}}}` is the following:

Let :math:`i_1(x), \\cdots, i_{n_{\\text{samples}}}(x)` be
the indices of the input training points
sorted by distance to the prediction point :math:`x`,
i.e.

.. math::

    \\|x-x_{i_1(x)}\\| \\leq \\cdots \\leq
    \\|x-x_{i_{n_{\\text{samples}}}(x)}\\|.

The ordered indices may be formally determined through the inductive formula

.. math::

    i_p(x) = \\underset{i\\in I_p(x)}{\\operatorname{argmin}}\\|x-x_i\\|,\\quad
    p=1,\\cdots,n_{\\text{samples}}

where

.. math::

    I_1(x) = \\{1,\\cdots,n_{\\text{samples}}\\}\\\\
    I_{p+1} = I_p(x)\\setminus \\{i_p(x)\\},\\quad
    p=1,\\cdots,n_{\\text{samples}}-1,

that is

.. math::

    I_p(x) = \\{1,\\cdots,n_{\\text{samples}}\\}\\setminus
    \\{i_1(x),\\cdots,i_{p-1}(x)\\}.

Then,
by denoting :math:`\\operatorname{mode}(\\cdot)` the mode operator,
i.e. the operator that extracts the element with the highest occurrence,
we may define the prediction operator as the mode of the set of output classes
associated to the :math:`k` first indices
(classes of the :math:`k`-nearest neighbors of :math:`x`):

.. math::

    f(x) = \\operatorname{mode}(y_{i_1(x)}, \\cdots, y_{i_k(x)})


This concept is implemented through the :class:`.KNNClassifier` class which
inherits from the :class:`.MLClassificationAlgo` class.

Dependence
----------
The classifier relies on the KNeighborsClassifier class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.neighbors.KNeighborsClassifier.html>`_.
"""
from __future__ import division, unicode_literals

import logging
from typing import Iterable, Optional, Union

from numpy import ndarray, stack
from sklearn.neighbors import KNeighborsClassifier

from gemseo.core.dataset import Dataset
from gemseo.mlearning.classification.classification import MLClassificationAlgo
from gemseo.mlearning.core.ml_algo import TransformerType

LOGGER = logging.getLogger(__name__)


class KNNClassifier(MLClassificationAlgo):
    """The k-nearest neighbors classification algorithm."""

    LIBRARY = "scikit-learn"
    ABBR = "KNN"

    def __init__(
        self,
        data,  # type: Dataset
        transformer=None,  # type: Optional[TransformerType]
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        n_neighbors=5,  # type: int
        **parameters  # type: Union[int,str]
    ):  # type: (...) -> None
        """
        Args:
            n_neighbors: The number of neighbors.
        """
        super(KNNClassifier, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            n_neighbors=n_neighbors,
            **parameters
        )
        self.algo = KNeighborsClassifier(n_neighbors, **parameters)

    def _fit(
        self,
        input_data,  # type:ndarray
        output_data,  # type:ndarray
    ):  # type: (...) -> None
        if output_data.shape[1] == 1:
            output_data = output_data.ravel()
        self.algo.fit(input_data, output_data)

    def _predict(
        self,
        input_data,  # type:ndarray
    ):  # type: (...) -> ndarray
        output_data = self.algo.predict(input_data).astype(int)
        if len(output_data.shape) == 1:
            output_data = output_data[:, None]
        return output_data

    def _predict_proba_soft(
        self,
        input_data,  # type: ndarray
    ):  # type: (...)-> ndarray
        probas = self.algo.predict_proba(input_data)
        if len(probas[0].shape) == 1:
            probas = probas[..., None]
        else:
            probas = stack(probas, axis=-1)
        return probas



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The mixture of experts for regression.

The mixture of experts (MoE) regression model expresses the output
as a weighted sum of local surrogate models,
where the weights are indicating the class of the input.

Inputs are grouped into clusters by a classification model
that is trained on a training set
where the output labels are determined through a clustering algorithm.
The outputs may be preprocessed
through a sensor or a dimension reduction algorithm.

The classification may either be hard,
in which case only one of the weights is equal to one,
and the rest equal to zero:

.. math::

    y = \\sum_{k=1}^K i_{C_k}(x) f_k(x),

or soft,
in which case the weights express the probabilities of belonging to each class:

.. math::

    y = \\sum_{k=1}^K \\mathbb{P}(x\\in C_k) f_k(x),

where
:math:`x` is the input,
:math:`y` is the output,
:math:`K` is the number of classes,
:math:`(C_k)_{k=1,\\cdots,K}` are the input spaces associated to the classes,
:math:`i_{C_k}(x)` is the indicator of class :math:`k`,
:math:`\\mathbb{P}(x\\in C_k)` is the probability of class :math:`k`
given :math:`x` and
:math:`f_k(x)` is the local surrogate model on class :math:`k`.

This concept is implemented through the :class:`.MixtureOfExperts` class
which inherits from the :class:`.MLRegressionAlgo` class.
"""
from __future__ import division, unicode_literals

import logging
from os.path import join
from typing import Callable, Dict, Iterable, List, NoReturn, Optional, Union

from numpy import ndarray, nonzero, unique, where, zeros

from gemseo.algos.design_space import DesignSpace
from gemseo.core.dataset import Dataset
from gemseo.mlearning.classification.factory import ClassificationModelFactory
from gemseo.mlearning.cluster.factory import ClusteringModelFactory
from gemseo.mlearning.core.ml_algo import DataType, MLAlgoParameterType, TransformerType
from gemseo.mlearning.core.selection import MLAlgoSelection
from gemseo.mlearning.core.supervised import SavedObjectType
from gemseo.mlearning.qual_measure.f1_measure import F1Measure
from gemseo.mlearning.qual_measure.mse_measure import MSEMeasure
from gemseo.mlearning.qual_measure.quality_measure import MLQualityMeasure
from gemseo.mlearning.qual_measure.quality_measure import OptionType as EvalOptionType
from gemseo.mlearning.qual_measure.silhouette import SilhouetteMeasure
from gemseo.mlearning.regression.factory import RegressionModelFactory
from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.string_tools import MultiLineString

LOGGER = logging.getLogger(__name__)

SavedObjectType = Union[SavedObjectType, str, Dict]

MLAlgoType = Dict[
    str,
    Optional[
        Union[str, DesignSpace, Dict[str, Union[str, int]], List[MLAlgoParameterType]]
    ],
]


class MixtureOfExperts(MLRegressionAlgo):
    """Mixture of experts regression.

    Attributes:
        hard (bool): Whether clustering/classification should be hard or soft.
        cluster_algo (str): The name of the clustering algorithm.
        classif_algo (str): The name of the classification algorithm.
        regress_algo (str): The name of the regression algorithm.
        cluster_params (Optional[MLAlgoParameterType]): The parameters
            of the clustering algorithm.
        classif_params (Optional[MLAlgoParameterType]): The parameters
            of the classification algorithm.
        regress_params (Optional[MLAlgoParameterType]): The parameters
            of the regression algorithm.
        cluster_measure (Dict[str,Union[str,EvalOptionType]]): The quality measure
            for the clustering algorithms.
        classif_measure (Dict[str,Union[str,EvalOptionType]]): The quality measure
            for the classification algorithms.
        regress_measure (Dict[str,Union[str,EvalOptionType]]): The quality measure
            for the regression algorithms.
        cluster_cands (List[MLAlgoType]): The clustering algorithm candidates.
        classif_cands (List[MLAlgoType]): The classification algorithm candidates.
        regress_cands (List[MLAlgoType]): The regression algorithm candidates.
        clusterer (MLClusteringAlgo): The clustering algorithm.
        classifier (MLClassificationAlgo): The classification algorithm.
        regress_models (List(MLRegressionAlgo)): The regression algorithms.
    """

    ABBR = "MoE"

    LABELS = "labels"

    _LOCAL_INPUT = "input"
    _LOCAL_OUTPUT = "output"

    def __init__(
        self,
        data,  # type: Dataset
        transformer=None,  # type: Optional[TransformerType]
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        hard=True,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            hard: Whether clustering/classification should be hard or soft.
        """
        super(MixtureOfExperts, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            hard=hard,
        )
        self.hard = hard
        self.cluster_algo = "KMeans"
        self.classif_algo = "KNNClassifier"
        self.regress_algo = "LinearRegression"
        self.cluster_params = {}
        self.classif_params = {}
        self.regress_params = {}

        self.cluster_measure = None
        self.classif_measure = None
        self.regress_measure = None

        self.set_clustering_measure(SilhouetteMeasure)
        self.set_classification_measure(F1Measure)
        self.set_regression_measure(MSEMeasure)

        self.cluster_cands = []
        self.regress_cands = []
        self.classif_cands = []

        self.clusterer = None
        self.classifier = None
        self.regress_models = None

    class DataFormatters(MLRegressionAlgo.DataFormatters):
        """Machine learning regression model decorators."""

        @classmethod
        def format_predict_class_dict(
            cls,
            predict,  # type: Callable[[ndarray],ndarray]
        ):  # type: (...) -> Callable[[DataType],DataType]
            """Make an array-based function be called with a dictionary of NumPy arrays.

            Args:
                predict: The function to be called;
                    it takes a NumPy array in input and returns a NumPy array.

            Returns:
                A function making the function 'predict' work with
                either a NumPy data array
                or a dictionary of NumPy data arrays indexed by variables names.
                The evaluation will have the same type as the input data.
            """

            def wrapper(
                self,
                input_data,  # type: DataType
                *args,
                **kwargs
            ):  # type: (...) -> DataType
                """Evaluate 'predict' with either array or dictionary-based input data.

                Firstly,
                the pre-processing stage converts the input data to a NumPy data array,
                if these data are expressed as a dictionary of NumPy data arrays.

                Then,
                the processing evaluates the function 'predict'
                from this NumPy input data array.

                Lastly,
                the post-processing transforms the output data
                to a dictionary of output NumPy data array
                if the input data were passed as a dictionary of NumPy data arrays.

                Args:
                    input_data: The input data.
                    *args: The positional arguments of the function 'predict'.
                    **kwargs: The keyword arguments of the function 'predict'.

                Returns:
                    The output data with the same type as the input one.
                """
                as_dict = isinstance(input_data, dict)
                if as_dict:
                    input_data = DataConversion.dict_to_array(
                        input_data, self.input_names
                    )
                output_data = predict(self, input_data, *args, **kwargs)
                if as_dict:
                    output_data = {self.LABELS: output_data}
                return output_data

            return wrapper

    def set_clusterer(
        self,
        cluster_algo,  # type: str
        **cluster_params  # type:Optional[MLAlgoParameterType]
    ):  # type: (...) -> None
        """Set the clustering algorithm.

        Args:
            cluster_algo: The name of the clustering algorithm.
            **cluster_params: The parameters of the clustering algorithm.
        """
        self.cluster_algo = cluster_algo
        self.cluster_params = cluster_params

    def set_classifier(
        self,
        classif_algo,  # type: str
        **classif_params  # type:Optional[MLAlgoParameterType]
    ):  # type: (...) -> None
        """Set the classification algorithm.

        Args:
            classif_algo: The name of the classification algorithm.
            **classif_params: The parameters of the classification algorithm.
        """
        self.classif_algo = classif_algo
        self.classif_params = classif_params

    def set_regressor(
        self,
        regress_algo,  # type: str
        **regress_params  # type:Optional[MLAlgoParameterType]
    ):  # type: (...) -> None
        """Set the regression algorithm.

        Args:
            regress_algo: The name of the regression algorithm.
            **regress_params: The parameters of the regression algorithm.
        """
        self.regress_algo = regress_algo
        self.regress_params = regress_params

    def set_clustering_measure(
        self,
        measure,  # type: MLQualityMeasure
        **eval_options  # type: EvalOptionType
    ):  # type: (...) -> None
        """Set the quality measure for the clustering algorithms.

        Args:
            measure: The quality measure.
            **eval_options: The options for the quality measure.
        """
        self.cluster_measure = {
            "measure": measure,
            "options": eval_options,
        }

    def set_classification_measure(
        self,
        measure,  # type: MLQualityMeasure
        **eval_options  # type: EvalOptionType
    ):  # type: (...) -> None
        """Set the quality measure for the classification algorithms.

        Args:
            measure: The quality measure.
            **eval_options: The options for the quality measure.
        """
        self.classif_measure = {
            "measure": measure,
            "options": eval_options,
        }

    def set_regression_measure(
        self,
        measure,  # type: MLQualityMeasure
        **eval_options  # type: EvalOptionType
    ):  # type: (...) -> None
        """Set the quality measure for the regression algorithms.

        Args:
            measure: The quality measure.
            **eval_options: The options for the quality measure.
        """
        self.regress_measure = {
            "measure": measure,
            "options": eval_options,
        }

    def add_clusterer_candidate(
        self,
        name,  # type: str
        calib_space=None,  # type: Optional[DesignSpace]
        calib_algo=None,  # type: Optional[Dict[str,Union[str,int]]]
        **option_lists  # type:Optional[List[MLAlgoParameterType]]
    ):  # type: (...) -> None
        """Add a candidate for clustering.

        Args:
            name: The name of a clustering algorithm.
            calib_space: The space
                defining the calibration variables.
            calib_algo: The name and options of the DOE or optimization
                algorithm, e.g. {"algo": "fullfact", "n_samples": 10}).
                If None, do not perform calibration.
            *** option_lists: Parameters for the clustering algorithm candidate.
                Each parameter has to be enclosed within a list.
                The list may contain different values to try out for the given
                parameter, or only one.
        """
        self.cluster_cands.append(
            dict(
                name=name,
                calib_space=calib_space,
                calib_algo=calib_algo,
                **option_lists
            )
        )

    def add_classifier_candidate(
        self,
        name,  # type: str
        calib_space=None,  # type: Optional[DesignSpace]
        calib_algo=None,  # type: Optional[Dict[str,Union[str,int]]]
        **option_lists  # type:Optional[List[MLAlgoParameterType]]
    ):  # type: (...) -> None
        """Add a candidate for classification.

        Args:
            name: The name of a classification algorithm.
            calib_space: The space
                defining the calibration variables.
            calib_algo: The name and options of the DOE or optimization
                algorithm, e.g. {"algo": "fullfact", "n_samples": 10}).
                If None, do not perform calibration.
            *** option_lists: Parameters for the clustering algorithm candidate.
                Each parameter has to be enclosed within a list.
                The list may contain different values to try out for the given
                parameter, or only one.
        """
        self.classif_cands.append(
            dict(
                name=name,
                calib_space=calib_space,
                calib_algo=calib_algo,
                **option_lists
            )
        )

    def add_regressor_candidate(
        self,
        name,  # type: str
        calib_space=None,  # type: Optional[DesignSpace]
        calib_algo=None,  # type: Optional[Dict[str,Union[str,int]]]
        **option_lists  # type:Optional[List[MLAlgoParameterType]]
    ):  # type: (...) -> None
        """Add a candidate for regression.

        Args:
            name: The name of a regression algorithm.
            calib_space: The space
                defining the calibration variables.
            calib_algo: The name and options of the DOE or optimization
                algorithm, e.g. {"algo": "fullfact", "n_samples": 10}).
                If None, do not perform calibration.
            *** option_lists: Parameters for the clustering algorithm candidate.
                Each parameter has to be enclosed within a list.
                The list may contain different values to try out for the given
                parameter, or only one.
        """
        self.regress_cands.append(
            dict(
                name=name,
                calib_space=calib_space,
                calib_algo=calib_algo,
                **option_lists
            )
        )

    @DataFormatters.format_predict_class_dict
    @DataFormatters.format_samples
    @DataFormatters.format_transform(transform_outputs=False)
    def predict_class(
        self,
        input_data,  # type: DataType
    ):  # type: (...) -> Union[int,ndarray]
        """Predict classes from input data.

        The user can specify these input data either as a NumPy array,
        e.g. :code:`array([1., 2., 3.])`
        or as a dictionary,
        e.g.  :code:`{'a': array([1.]), 'b': array([2., 3.])}`.

        The output data type will be consistent with the input data type.

        Args:
            input_data: The input data.

        Returns:
            The predicted classes.
        """
        return self.classifier.predict(input_data)

    @DataFormatters.format_input_output
    def predict_local_model(
        self,
        input_data,  # type: DataType
        index,  # type: int
    ):  # type: (...) -> DataType
        """Predict output data from input data.

        The user can specify these input data either as a NumPy array,
        e.g. :code:`array([1., 2., 3.])`
        or as a dictionary,
        e.g.  :code:`{'a': array([1.]), 'b': array([2., 3.])}`.

        The output data type will be consistent with the input data type.

        Args:
            input_data: The input data.
            index: The index of the local regression model.

        Returns:
            The predicted output data.
        """
        return self.regress_models[index].predict(input_data)

    def _fit(
        self,
        input_data,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> None
        dataset = Dataset("training_set")
        dataset.add_group(
            Dataset.INPUT_GROUP,
            input_data,
            [self._LOCAL_INPUT],
            {self._LOCAL_INPUT: input_data.shape[1]},
        )
        dataset.add_group(
            Dataset.OUTPUT_GROUP,
            output_data,
            [self._LOCAL_OUTPUT],
            {self._LOCAL_OUTPUT: output_data.shape[1]},
            cache_as_input=False,
        )
        self._fit_clusters(dataset)
        self._fit_classifier(dataset)
        self._fit_regressors(dataset)

    def _fit_clusters(
        self,
        dataset,  # type:Dataset
    ):  # type: (...) -> None
        """Train the clustering algorithm.

        The methods adds resulting labels as a new output in the dataset.

        Args:
            dataset: The dataset containing input and output data.
        """
        if not self.cluster_cands:
            factory = ClusteringModelFactory()
            self.clusterer = factory.create(
                self.cluster_algo, data=dataset, **self.cluster_params
            )
            self.clusterer.learn()
        else:
            selector = MLAlgoSelection(
                dataset,
                self.cluster_measure["measure"],
                **self.cluster_measure["options"]
            )
            for cand in self.cluster_cands:
                selector.add_candidate(**cand)
            self.clusterer = selector.select()
            LOGGER.info("Selected clusterer:")
            with MultiLineString.offset():
                LOGGER.info("%s", self.clusterer)

        labels = self.clusterer.labels[:, None]
        dataset.add_variable(self.LABELS, labels, self.LABELS, False)

    def _fit_classifier(
        self,
        dataset,  # type:Dataset
    ):  # type: (...) -> None
        """Train the classification algorithm.

        Args:
            dataset: The dataset containing labeled input and output data.
        """
        if not self.classif_cands:
            factory = ClassificationModelFactory()
            self.classifier = factory.create(
                self.classif_algo,
                data=dataset,
                output_names=[self.LABELS],
                **self.classif_params
            )
            self.classifier.learn()
        else:
            selector = MLAlgoSelection(
                dataset,
                self.classif_measure["measure"],
                **self.classif_measure["options"]
            )
            for cand in self.classif_cands:
                selector.add_candidate(output_names=[[self.LABELS]], **cand)
            self.classifier = selector.select()
            LOGGER.info("Selected classifier:")
            with MultiLineString.offset():
                LOGGER.info("%s", self.classifier)

    def _fit_regressors(
        self,
        dataset,  # type:Dataset
    ):  # type: (...) -> None
        """Train the local regression models on each cluster separately.

        Args:
            dataset: The dataset containing labeled input and output data.
        """
        factory = RegressionModelFactory()
        self.regress_models = []
        for index in range(self.clusterer.n_clusters):
            samples = nonzero(self.clusterer.labels == index)[0]
            if not self.regress_cands:
                local_model = factory.create(
                    self.regress_algo, data=dataset, **self.regress_params
                )
                local_model.learn(samples=samples)
            else:
                selector = MLAlgoSelection(
                    dataset,
                    self.regress_measure["measure"],
                    samples=samples,
                    **self.regress_measure["options"]
                )
                for cand in self.regress_cands:
                    selector.add_candidate(**cand)
                local_model = selector.select()
                LOGGER.info("Selected regressor for cluster %s:", index)
                with MultiLineString.offset():
                    LOGGER.info("%s", local_model)

            self.regress_models.append(local_model)

    def _predict_all(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Predict output of each regression model for given input data.

        This method stacks the different outputs along a new axis.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).

        Returns:
            The output data with shape (n_samples, n_clusters, n_outputs).
        """
        # dim(input_data)  = (n_samples, n_inputs)
        # dim(output_data) = (n_samples, n_clusters, n_outputs)
        output_data = zeros(
            (input_data.shape[0], self.n_clusters, self.regress_models[0].output_shape)
        )
        for i in range(self.n_clusters):
            output_data[:, i] = self.regress_models[i].predict(input_data)
        return output_data

    def _predict(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        # dim(probas)         = (n_samples, n_clusters,     1    )
        # dim(local_outputs)  = (n_samples, n_clusters, n_outputs)
        # dim(contributions)  = (n_samples, n_clusters, n_outputs)
        # dim(global_outputs) = (n_samples, n_outputs)
        probas = self.classifier.predict_proba(input_data, hard=self.hard)
        local_outputs = self._predict_all(input_data)
        contributions = probas * local_outputs
        global_outputs = contributions.sum(axis=1)
        return global_outputs

    def _predict_jacobian(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        if self.hard:
            jacobians = self._predict_jacobian_hard(input_data)
        else:
            jacobians = self._predict_jacobian_soft(input_data)
        return jacobians

    def _predict_jacobian_hard(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Predict the Jacobian matrices of the regression model at input_data.

        This method uses a hard classification.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).

        Returns:
            The predicted Jacobian data with shape (n_samples, n_outputs, n_inputs).
        """
        n_samples = input_data.shape[0]
        classes = self.classifier.predict(input_data)[..., 0]
        unq_classes = unique(classes)
        jacobians = zeros(
            (
                n_samples,
                self.regress_models[0].output_shape,
                self.regress_models[0].input_shape,
            )
        )
        for klass in unq_classes:
            inds_kls = where(classes == klass)
            jacobians[inds_kls] = self.regress_models[klass].predict_jacobian(
                input_data[inds_kls]
            )
        return jacobians

    def _predict_jacobian_soft(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> NoReturn
        """Predict the Jacobian matrices of the regression model at input_data.

        This method uses a soft classification.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).

        Returns:
            The predicted Jacobian data with shape (n_samples, n_outputs, n_inputs).
        """
        raise NotImplementedError

    def _save_algo(
        self,
        directory,  # type: str
    ):  # type: (...) -> None
        self.clusterer.save(join(directory, "clusterer"))
        self.classifier.save(join(directory, "classifier"))
        for i, local_model in enumerate(self.regress_models):
            local_model.save(join(directory, "local_model_{}".format(i)))

    def load_algo(
        self,
        directory,  # type: str
    ):  # type: (...) -> None
        cluster_factory = ClusteringModelFactory()
        classif_factory = ClassificationModelFactory()
        regress_factory = RegressionModelFactory()
        self.clusterer = cluster_factory.load(join(directory, "clusterer"))
        self.classifier = classif_factory.load(join(directory, "classifier"))
        self.regress_models = []
        for i in range(self.n_clusters):
            self.regress_models.append(
                regress_factory.load(join(directory, "local_model_{}".format(i)))
            )

    def __str__(self):  # type: (...) -> None
        string = MultiLineString()
        string.add(super(MixtureOfExperts, self).__str__())
        string.indent()
        string.indent()
        string.add("Clustering")
        string.indent()
        string.add(str(self.clusterer).split("\n")[0])
        string.dedent()
        string.add("Classification")
        string.indent()
        string.add(str(self.classifier).split("\n")[0])
        string.dedent()
        string.add("Regression")
        string.indent()
        for i, local_model in enumerate(self.regress_models):
            string.add("Local model {}", i)
            string.indent()
            string.add(str(local_model).split("\n")[0])
            string.dedent()
        return str(string)

    def _get_objects_to_save(self):  # type: (...) -> Dict[str,SavedObjectType]
        objects = super(MixtureOfExperts, self)._get_objects_to_save()
        objects["cluster_algo"] = self.cluster_algo
        objects["classif_algo"] = self.classif_algo
        objects["regress_algo"] = self.regress_algo
        objects["cluster_params"] = self.cluster_params
        objects["classif_params"] = self.classif_params
        objects["regress_params"] = self.regress_params
        return objects

    @property
    def labels(self):  # type:(...) -> List[int]
        """The cluster labels."""
        return self.clusterer.labels

    @property
    def n_clusters(self):  # type:(...) -> int
        """The number of clusters."""
        return self.clusterer.n_clusters



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
"""
Build a diagonal DOE for scalable model construction
****************************************************
"""
from __future__ import division, unicode_literals

import logging

from numpy import array

from gemseo.algos.doe.doe_lib import DOELibrary

LOGGER = logging.getLogger(__name__)


class DiagonalDOE(DOELibrary):

    """Class used for creation of a diagonal DOE."""

    ALGO_LIST = ["DiagonalDOE"]
    ALGO_DESC = {"DiagonalDOE": "Diagonal design of experiments"}

    def __init__(self):
        """Constructor, initializes the DOE samples."""
        super(DiagonalDOE, self).__init__()

        for algo in self.ALGO_LIST:
            description = DiagonalDOE.ALGO_DESC[algo]
            self.lib_dict[algo] = {
                DOELibrary.LIB: self.__class__.__name__,
                DOELibrary.INTERNAL_NAME: algo,
                DOELibrary.DESCRIPTION: description,
            }

    def _get_options(
        self,
        eval_jac=False,
        n_processes=1,
        wait_time_between_samples=0.0,
        n_samples=1,
        reverse=None,
        max_time=0,
        **kwargs
    ):  # pylint: disable=W0221
        """Sets the options.

        :param eval_jac: evaluate jacobian
        :type eval_jac: bool
        :param n_processes: number of processes
        :type n_processes: int
        :param wait_time_between_samples: waiting time between two samples
        :type wait_time_between_samples: float
        :param n_samples: number of samples
        :type n_samples: int
        :param reverse: list of dimensions or variables to sample from their
            upper bounds to their lower bounds. Default: None.
        :type reverse: list(str)
        :param max_time: maximum runtime in seconds,
            disabled if 0 (Default value = 0)
        :type max_time: float
        :param kwargs: additional arguments
        """
        wtbs = wait_time_between_samples
        return self._process_options(
            eval_jac=eval_jac,
            n_processes=n_processes,
            wait_time_between_samples=wtbs,
            n_samples=n_samples,
            reverse=reverse,
            max_time=max_time,
            **kwargs
        )

    def _generate_samples(self, **options):
        """Generates the list of x samples.

        :param options: the options dict for the algorithm,
            see associated JSON file
        """
        reverse = options.get("reverse", [])
        names = self.problem.design_space.variables_names
        sizes = self.problem.design_space.variables_sizes
        name_by_index = {}
        start = 0
        for name in names:
            for index in range(start, start + sizes[name]):
                name_by_index[index] = name
            start += sizes[name]
        samples = []
        for index in range(self.problem.dimension):
            if str(index) in reverse or name_by_index[index] in reverse:
                samples.append(
                    [
                        point / (options[self.N_SAMPLES] - 1.0)
                        for point in range(options[self.N_SAMPLES] - 1, -1, -1)
                    ]
                )
            else:
                samples.append(
                    [
                        point / (options[self.N_SAMPLES] - 1.0)
                        for point in range(0, options[self.N_SAMPLES])
                    ]
                )
        samples = array(samples).T
        return samples



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A transformer to apply operations on NumPy arrays.

The abstract :class:`.Transformer` class implements the concept of a data transformer.
Inheriting classes shall implement the :meth:`.Transformer.fit`,
:meth:`.Transformer.transform`
and possibly :meth:`.Transformer.inverse_transform` methods.

.. seealso::

   :mod:`~gemseo.mlearning.transform.scaler.scaler`
   :mod:`~gemseo.mlearning.transform.dimension_reduction.dimension_reduction`
"""
from __future__ import division, unicode_literals

from typing import NoReturn, Optional, Union

import six
from custom_inherit import DocInheritMeta
from numpy import ndarray

TransformerFitOptionType = Union[float, int, str]


@six.add_metaclass(
    DocInheritMeta(
        abstract_base_class=True,
        style="google_with_merge",
    )
)
class Transformer(object):
    """Transformer baseclass.

    Attributes:
        name (str): The name of the transformer.
        parameters (str): The parameters of the transformer.
    """

    CROSSED = False

    def __init__(
        self,
        name="Transformer",  # type: str
        **parameters  # type: Optional[Union[float,int,str,bool]]
    ):  # type: (...) -> None
        """
        Args:
            name: A name for this transformer.
            **parameters: The parameters of the transformer.
        """
        self.name = name
        self.parameters = parameters

    def duplicate(self):  # type: (...) -> Transformer
        """Duplicate the current object.

        Returns:
            A deepcopy of the current instance.
        """
        return self.__class__(self.name, **self.parameters)

    def fit(
        self,
        data,  # type: ndarray
        *args  # type: TransformerFitOptionType
    ):  # type: (...) -> NoReturn
        """Fit the transformer to the data.

        Args:
            data: The data to be fitted.
        """
        raise NotImplementedError

    def transform(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> NoReturn
        """Transform the data.

        Args:
            data: The data to be transformed.

        Returns:
            The transformed data.
        """
        raise NotImplementedError

    def inverse_transform(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> NoReturn
        """Perform an inverse transform on the data.

        Args:
            data: The data to be inverse transformed.

        Returns:
            The inverse transformed data.
        """
        raise NotImplementedError

    def fit_transform(
        self,
        data,  # type: ndarray
        *args  # type: TransformerFitOptionType
    ):  # type: (...) -> ndarray
        """Fit the transformer to the data and transform the data.

        Args:
            data: The data to be transformed.

        Returns:
            The transformed data.
        """
        self.fit(data, *args)
        return self.transform(data)

    def compute_jacobian(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> NoReturn
        """Compute Jacobian of transformer.transform().

        Args:
            data: The data where the Jacobian is to be computed.

        Returns:
            The Jacobian matrix.
        """
        raise NotImplementedError

    def compute_jacobian_inverse(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> NoReturn
        """Compute Jacobian of the transformer.inverse_transform().

        Args:
            data: The data where the Jacobian is to be computed.

        Returns:
            The Jacobian matrix.
        """
        raise NotImplementedError

    def __str__(self):  # type: (...) -> str
        return self.__class__.__name__



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Pierre-Jean Barjhoux
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A radar plot of constraints."""
from __future__ import division, unicode_literals

import logging
from typing import Sequence

from numpy import vstack, zeros

from gemseo.core.dataset import Dataset
from gemseo.post.dataset.radar_chart import RadarChart as RadarChartPost
from gemseo.post.opt_post_processor import OptPostProcessor

LOGGER = logging.getLogger(__name__)


class RadarChart(OptPostProcessor):
    """Plot on radar style chart a list of constraint functions.

    This class has the responsability of plotting on radar style chart a list
    of constraint functions at a given iteration.

    By default, the iteration is the last one.
    It is possible either to save the plot, to show the plot or both.
    """

    def _plot(
        self,
        constraints_list,  # type: Sequence[str]
        iteration=-1,  # type: int
        figsize_x=8,  # type: int
        figsize_y=8,  # type: int
    ):  # type: (...) -> None
        """
        Args:
            constraints_list: The names of the constraints.
            iteration: The number of iteration to post-process.
            figsize_x: The size of the figure in horizontal direction (inches).
            figsize_y: The size of the figure in vertical direction (inches).
        """
        # retrieve the constraints values
        add_dv = False
        all_constr_names = self.opt_problem.get_constraints_names()

        for func in constraints_list:
            if func not in all_constr_names:
                raise ValueError(
                    "Cannot build radar chart; "
                    "function {} is not among constraints names"
                    " or does not exist.".format(func)
                )

        cstr_values, cstr_names, _ = self.database.get_history_array(
            constraints_list, add_dv=add_dv
        )

        if iteration < -1 or iteration >= len(self.database):
            raise ValueError(
                "iteration should be either equal to -1 or positive and lower than "
                "maximum iteration = {}".format(len(self.database))
            )
        cstr_values = cstr_values[iteration, :].ravel()

        dataset = Dataset("Constraints")
        values = vstack((cstr_values, zeros(len(cstr_values))))
        dataset.add_group(
            dataset.DEFAULT_GROUP,
            values,
            cstr_names,
            {name: 1 for name in cstr_names},
        )
        dataset.row_names = ["computed constraints", "limit constraint"]

        radar = RadarChartPost(dataset)
        radar.linestyle = {"computed constraints": "-", "limit constraint": "--"}
        radar.color = {"computed constraints": "k", "limit constraint": "r"}
        if iteration == -1:
            radar.title = "Constraints at last iteration"
        else:
            radar.title = "Constraints at iteration {}".format(iteration)
        radar.figsize_x = figsize_x
        radar.figsize_y = figsize_y
        figures = radar.execute(save=False, show=False, display_zero=False)
        for figure in figures:
            self._add_figure(figure)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

r"""
Scalable module from Tedford and Martins (2010)
***********************************************

The modules located in this directory offer a set of
classes relative to the scalable problem introduced in the paper:

    Tedford NP, Martins JRRA (2010), Benchmarking
    multidisciplinary design optimization algorithms,
    Optimization and Engineering, 11(1):159-183.

Overview
~~~~~~~~

This scalable problem aims to minimize an objective function
quadratically depending on shared design parameters and coupling variables,
under inequality constraints linearly depending on these coupling variables.

System discipline
-----------------

A system discipline computes the constraints and the objective
in function of the shared design parameters and coupling variables.

The discipline takes the global design parameters :math:`z` and
the coupling variables :math:`y_1,y_2,\ldots,y_N` as inputs
and returns the objective function value :math:`f(x,y(x,y))` to minimize as
well as the inequality constraints ones
:math:`c_1(y_1),c_2(y_2),\ldots,c_N(y_N)` which are expressed as:

.. math::

   f(z,y) = |z|_2^2 + \sum_{i=1}^N |y_i|_2^2


and:

.. math::

   c_i(y_i) = 1- C_i^{-T}Iy_i

Strongly coupled disciplines
----------------------------

The coupling variables are the outputs of strongly coupled disciplines.

Each strongly coupled discipline computes a set of coupling variables
linearly depending on local design parameters, shared design parameters,
coupling variables from other strongly coupled disciplines,
and belonging to the unit hypercube.

The i-th discipline takes local design parameters :math:`x_i`
and shared design parameters :math:`z` in input as well as coupling
variables :math:`\left(y_i\right)_{1\leq j \leq N\atop j\neq i}`
from :math:`N-1` elementary disciplines,
and returns the coupling variables:

.. math::

    y_i =\frac{\tilde{y}_i+C_{z,i}.1+C_{x_i}.1}{\sum_{j=1 \atop j
    \neq i}^NC_{y_j,i}.1+C_{z,i}.1+C_{x_i}.1} \in [0,1]^{n_{y_i}}

where:

.. math::

    \tilde{y}_i = - C_{z,i}.z - C_{x_i}.x_i +
    \sum_{j=1 \atop j \neq i}^N C_{y_j,i}.y_j

Scalability
-----------

This problem is said "scalable"
because several sizing features can be chosen by the user:

- the number of local design parameters for each discipline,
- the number of shared design parameters,
- the number of coupling variables for each discipline,
- the number of disciplines.

A given sizing configuration is called "scaling strategy"
and this scalable module is particularly useful to compare different MDO
formulations with respect to the scaling strategy.
"""



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Draw surfaces from a :class:`.Dataset`.

A :class:`.Surfaces` plot represents samples
of a functional variable :math:`z(x,y)` discretized over a 2D mesh.
Both evaluations of :math:`z` and mesh are stored in a :class:`.Dataset`,
:math:`z` as a parameter and the mesh as a metadata.
"""
from __future__ import division, unicode_literals

from typing import List, Mapping, Optional, Sequence

import matplotlib.pyplot as plt
import matplotlib.tri as mtri
from matplotlib.figure import Figure

from gemseo.post.dataset.dataset_plot import DatasetPlot


class Surfaces(DatasetPlot):
    """Plot surfaces y_i over the mesh x."""

    def _plot(
        self,
        properties,  # type: Mapping
        mesh,  # type: str
        variable,  # type: str
        samples=None,  # type:Optional[Sequence[int]]
        add_points=False,  # type: bool
    ):  # type: (...) -> List[Figure]
        """
        Args:
            mesh: The name of the dataset metadata corresponding to the mesh.
            variable: The name of the variable for the x-axis.
            samples: The indices of the samples to plot. If None, plot all samples.
            add_points: If True then display the samples over the surface plot.
        """
        color = properties.get(self.COLOR) or "blue"
        colormap = properties.get(self.COLORMAP) or "Blues"
        x_data = self.dataset.metadata[mesh][:, 0]
        y_data = self.dataset.metadata[mesh][:, 1]
        if samples is not None:
            outputs = self.dataset[variable][variable][samples, :]
        else:
            outputs = self.dataset[variable][variable]

        sample = 0
        fig = []
        for z_data, variable_component in zip(outputs, self.dataset.row_names):
            fig.append(plt.figure())
            axes = fig[-1].add_subplot(1, 1, 1)
            triangle = mtri.Triangulation(x_data, y_data)
            tcf = axes.tricontourf(triangle, z_data, cmap=colormap)
            if add_points:
                axes.scatter(x_data, y_data, color=color)
            axes.set_title("{} - {}".format(variable, variable_component))
            fig[-1].colorbar(tcf)
            fig[-1] = plt.gcf()
            sample += 1
        return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""Draw Andrews curves from a :class:`.Dataset`.

The :class:`.AndrewsCurves` class implements the Andrew plot, a.k.a. Andrews curves,
which is a way to visualize :math:`n` samples of a high-dimensional vector

.. math::

   x=(x_1,x_2,\ldots,x_d)\in\mathbb{R}^d

in a 2D referential by projecting each sample

.. math::

   x^{(i)}=(x_1^{(i)},x_2^{(i)},\ldots,x_d^{(i)})

onto the vector

.. math::

   \left(\frac{1}{\sqrt{2}},\sin(t),\cos(t),\sin(2t),\cos(2t), \ldots\right)

which is composed of the :math:`d` first elements of the Fourier series:

.. math::

   f_i(t)=\left(\frac{x_1^{(i)}}{\sqrt{2}},x_2^{(i)}\sin(t),x_3^{(i)}\cos(t),
   x_4^{(i)}\sin(2t),x_5^{(i)}\cos(2t),\ldots\right)

Each curve :math:`t\mapsto f_i(t)` is plotted
over the interval :math:`[-\pi,\pi]`
and structure in the data may be visible in these :math:`n` Andrews curves.

A variable name can be passed to the :meth:`.DatasetPlot.execute` method
by means of the :code:`classifier` keyword
in order to color the curves according to the value of the variable name.
This is useful when the data is labeled.
"""
from __future__ import division, unicode_literals

from typing import List, Mapping

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from pandas.plotting import andrews_curves

from gemseo.post.dataset.dataset_plot import DatasetPlot, DatasetPlotPropertyType


class AndrewsCurves(DatasetPlot):
    """Andrews curves."""

    def _plot(
        self,
        properties,  # type: Mapping[str,DatasetPlotPropertyType]
        classifier,  # type: str
    ):  # type: (...) -> List[Figure]
        """
        Args:
            classifier: The name of the variable to group the data.
        """
        if classifier not in self.dataset.variables:
            raise ValueError(
                "Classifier must be one of these names: "
                + ", ".join(self.dataset.variables)
            )

        dataframe = self.dataset.export_to_dataframe()
        label, varname = self._get_label(classifier)
        if self.dataset.strings_encoding[label]:
            for comp, codes in self.dataset.strings_encoding[label].items():
                column = (self.dataset.get_group(label), label, str(comp))
                for key, value in codes.items():
                    dataframe.loc[dataframe[column] == key, column] = value
        andrews_curves(dataframe, varname)
        return [plt.gcf()]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Pierre-Jean Barjhoux
#       :author: Francois Gallard, integration and cleanup
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Implementation of the Lagrange multipliers
******************************************
"""
from __future__ import division, unicode_literals

import logging

from numpy import arange, array, atleast_2d, concatenate, where, zeros
from numpy.linalg import lstsq, matrix_rank, norm
from scipy.optimize import linprog

from gemseo.algos.design_space import DesignSpace
from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.third_party.prettytable import PrettyTable

LOGGER = logging.getLogger(__name__)


class LagrangeMultipliers(object):
    r"""
    Class that implements the computation of Lagrange Multipliers

    Denote :math:`x^\ast` an optimal solution of the optimization problem
    below.

    .. math::
        \begin{aligned}
            & \text{Minimize}    & & f(x) \\
            & \text{relative to} & & x \\
            & \text{subject to}  & & \left\{\begin{aligned}
                                                & g(x)\le0, \\
                                                & h(x)=0, \\
                                                & \ell\le x\le u.
                                     \end{aligned}\right.
        \end{aligned}

    If the constraints are qualified at :math:`x^\ast` then the Lagrange
    multipliers of :math:`x^\ast` are the vectors :math:`\lambda_g`,
    :math:`\lambda_h`, :math:`\lambda_\ell` and :math:`\lambda_u` satisfying

    .. math::
        \left\{\begin{aligned}
            &\frac{\partial f}{\partial x}(x^\ast)
            +\lambda_g^\top\frac{\partial g}{\partial x}(x^\ast)
            +\lambda_h^\top\frac{\partial h}{\partial x}(x^\ast)
            +\sum_j\lambda_{\ell,j}+\sum_j\lambda_{u,j}
            =0,\\
            &\lambda_{g,i}\ge0\text{ if }g_i(x^\ast)=0,
            \text{ otherwise }\lambda_{g,i}=0,\\
            &\lambda_{\ell,j}\le0\text{ if }x^\ast_j=\ell_j,
            \text{ otherwise }\lambda_{\ell,j}=0,\\
            &\lambda_{u,j}\ge0\text{ if }x^\ast_j=u_j,
            \text{ otherwise }\lambda_{u,j}=0.
        \end{aligned}\right.

    """
    LOWER_BOUNDS = "lower_bounds"
    UPPER_BOUNDS = "upper_bounds"
    INEQUALITY = "inequality"
    EQUALITY = "equality"
    CSTR_LABELS = [LOWER_BOUNDS, UPPER_BOUNDS, INEQUALITY, EQUALITY]

    def __init__(self, opt_problem):
        """Constructor.

        :param opt_problem: optimization problem on which Lagrange multipliers
            shall be computed
        """
        self._check_inputs(opt_problem)
        self.opt_problem = opt_problem
        self.active_lb_names = []
        self.active_ub_names = []
        self.active_ineq_names = []
        self.active_eq_names = []
        self.lagrange_multipliers = None
        self.__normalized = opt_problem.preprocess_options.get("normalize", False)

    @staticmethod
    def _check_inputs(opt_problem):
        """
        Checks inputs : verify that opt_problem is an
        instance of OptimizationProblem

        :param opt_problem: optimization problem on which Lagrange multipliers
            shall be computed
        """
        if not isinstance(opt_problem, OptimizationProblem):
            raise ValueError(
                "Argument of LagrangeMultiplier class"
                + " has to be an instance of OptimizationProblem"
            )
        if opt_problem.solution is None:
            raise ValueError("Optimization problem was not solved !")

    def compute(self, x_vect, ineq_tolerance=1e-6, rcond=-1):
        """Computes and returns the Lagrange multipliers, as a post-processing of the
        optimal point.

        This solves :
        (d ActiveConstraints)'               d Objective
        (-------------------)  . Lambda = -  -----------
        (d X                )                d X

        :param x_vect: x point on which the multipliers shall be computed
        :param ineq_tolerance: tolerance on inequality constraints
        :param rcond: float, optional
               Cut-off ratio for small singular values of the jacobian.
               see sipy.linalg.lsq
        """
        LOGGER.info("Computation of Lagrange multipliers")

        # Check feasibility
        self._check_feasibility(x_vect)

        # get jacobian of all active constraints, and an
        # ordered list of their name
        jac_act, _ = self._get_jac_act(x_vect, ineq_tolerance)
        if jac_act is None:
            # There is no active constraint
            multipliers = []
            self._store_multipliers(multipliers)
            return self.lagrange_multipliers
        lhs = jac_act.T
        act_constr_nb = lhs.shape[1]
        rank = matrix_rank(lhs)
        LOGGER.info("Found %s active constraints", str(act_constr_nb))
        LOGGER.info("Rank of jacobian = %s", str(rank))
        if act_constr_nb > rank:
            LOGGER.warning("Number of active constraints > rank !")

        # get jacobian of objective
        obj_jac = self._get_obj_jac(x_vect)
        rhs = -obj_jac.T

        # Compute the Lagrange multipliers as a feasible solution of a
        # linear optimization problem
        act_eq_constr_nb = len(self.active_eq_names)
        bounds = [(0, None)] * (act_constr_nb - act_eq_constr_nb) + [
            (None, None)
        ] * act_eq_constr_nb
        optim_result = linprog(zeros(act_constr_nb), A_eq=lhs, b_eq=rhs, bounds=bounds)
        if optim_result.status == 2:
            LOGGER.warning("Lagrange multipliers appear not to exist")
        if optim_result.success:
            mul = optim_result.x
        else:
            # If the linear optimization failed then obtain the Lagrange
            # multipliers as a solution of a least-square problem
            mul, residuals, _, sval = lstsq(lhs, rhs, rcond=rcond)
            LOGGER.info("Min singular values of jacobian = %s", str(sval.min()))
            LOGGER.info("Residuals norm = %s", str(norm(residuals)))

        # stores multipliers in a dictionary
        self._store_multipliers(mul)

        return self.lagrange_multipliers

    def _check_feasibility(self, x_vect):
        """Checks that the given point is in the design space and satisfies all
        constraints.

        :param x_vect: point at which the Lagrange multipliers are to be
            computed
        :type x_vect: ndarray
        """
        problem = self.opt_problem

        # Check that the point is within bounds
        problem.design_space.check_membership(x_vect)

        # Check that the point satisfies other constraints
        values, _ = problem.evaluate_functions(
            x_vect, eval_obj=False, normalize=False, no_db_no_norm=True
        )
        feasible = problem.is_point_feasible(values)
        if not feasible:
            LOGGER.warning("Infeasible point, Lagrange multipliers may not" " exist")

    def _get_act_bound_jac(self, act_bounds):
        """Returns the jacobian of active  bounds constraints sign is not taken into
        account (matrix is made of 0 and 1)"""
        dspace = self.opt_problem.design_space
        x_dim = dspace.dimension
        dim_act = sum(len(where(bnd)[0]) for bnd in act_bounds.values())
        if dim_act == 0:
            return None, []
        act_array = concatenate([act_bounds[var] for var in dspace.variables_names])

        bnd_jac = zeros((dim_act, x_dim))

        if self.__normalized:
            norm_factor = dspace.get_upper_bounds() - dspace.get_lower_bounds()
            act_jac = norm_factor[act_array]
        else:
            act_jac = 1.0
        bnd_jac[arange(dim_act), act_array] = act_jac
        indexed_varnames = array(dspace.get_indexed_variables_names())
        act_b_names = indexed_varnames[act_array].tolist()
        return bnd_jac, act_b_names

    def __get_act_ineq_jac(self, x_vect, ineq_tolerance=1e-6):
        """Returns the jacobian of active inequality constraints defined by user in the
        optimization problem."""
        # retrieves the active functions and the indices :
        # a function is active if at least
        # one of its component (in case of multidimensional constraints) is
        # active
        act_func = self.opt_problem.get_active_ineq_constraints(x_vect, ineq_tolerance)

        dspace = self.opt_problem.design_space

        if self.__normalized:
            x_vect = dspace.normalize_vect(x_vect)
        jac = []
        names = []

        for func, act_set in act_func.items():
            if True in act_set:
                ineq_jac = func.jac(x_vect)
                if len(ineq_jac.shape) == 1:
                    # Make sure the Jacobian is a 2-dimensional array
                    ineq_jac = ineq_jac.reshape((1, x_vect.size))
                else:
                    ineq_jac = ineq_jac[act_set, :]
                jac.append(ineq_jac)
                if func.dim == 1:
                    names.append(func.name)
                else:
                    names += [
                        self._get_component_name(func.name, i)
                        for i, active in enumerate(act_set)
                        if active
                    ]
        if jac:
            jac = concatenate(jac)
        else:
            jac = None
        return jac, names

    def _get_act_eq_jac(self, x_vect):
        """Returns jacobian of active equality constraints defined by user in the
        optimization problem."""
        eq_functions = self.opt_problem.get_eq_constraints()
        # loop on equality functions
        # NB: as the solution (x_vect) is supposed to be feasible,
        # all functions (on all dimensions) are supposed to be active
        jac = []
        names = []
        dspace = self.opt_problem.design_space

        if self.__normalized:
            x_vect = dspace.normalize_vect(x_vect)

        for eq_function in eq_functions:
            eq_jac = atleast_2d(eq_function.jac(x_vect))
            jac.append(eq_jac)
            if eq_function.dim == 1:
                names.append(eq_function.name)
            else:
                names += [
                    self._get_component_name(eq_function.name, i)
                    for i in range(eq_jac.shape[0])
                ]
        if jac:
            jac = concatenate(jac)
        else:
            jac = None
        return jac, names

    def _get_obj_jac(self, x_vect):
        """Returns objective jacobian."""
        if self.__normalized:
            x_vect = self.opt_problem.design_space.normalize_vect(x_vect)

        return self.opt_problem.objective.jac(x_vect)

    def _get_jac_act(self, x_vect, ineq_tolerance=1e-6):
        """Returns active constraints jacobian, and the name of each component of each
        function."""
        # Bounds jacobian
        dspace = self.opt_problem.design_space
        act_lb, act_ub = dspace.get_active_bounds(x_vect, tol=ineq_tolerance)
        lb_jac_act, self.active_lb_names = self._get_act_bound_jac(act_lb)
        if lb_jac_act is not None:
            lb_jac_act *= -1
        ub_jac_act, self.active_ub_names = self._get_act_bound_jac(act_ub)

        # inequality names
        tol = ineq_tolerance
        ineq_jac, self.active_ineq_names = self.__get_act_ineq_jac(x_vect, tol)
        # equality names
        eq_jac, eq_names_act = self._get_act_eq_jac(x_vect)
        self.active_eq_names = eq_names_act

        names_list = (
            self.active_lb_names
            + self.active_ub_names
            + self.active_ineq_names
            + eq_names_act
        )
        jac_list = [lb_jac_act, ub_jac_act, ineq_jac, eq_jac]
        jac_list = [jac for jac in jac_list if jac is not None]
        if jac_list:
            jac_act_arr = concatenate(jac_list, axis=0)
        else:
            # There no active constraint
            jac_act_arr = None

        return jac_act_arr, names_list

    def _store_multipliers(self, multipliers):
        """Stores multipliers in a dictionary."""
        lag = {}

        i_min = 0
        n_act = len(self.active_lb_names)
        if n_act > 0:
            l_b_mult = multipliers[i_min : i_min + n_act]
            lag[self.LOWER_BOUNDS] = (self.active_lb_names, l_b_mult)
            i_min += n_act
            wrong_inds = where(l_b_mult < 0.0)[0]
            if wrong_inds.size > 0:
                names_neg = array(self.active_lb_names)[wrong_inds]
                LOGGER.warning(
                    "Negative Lagrange multipliers for "
                    "lower bounds on variables"
                    "%s !",
                    str(names_neg),
                )
        n_act = len(self.active_ub_names)
        if n_act > 0:
            u_b_mult = multipliers[i_min : i_min + n_act]
            lag[self.UPPER_BOUNDS] = (self.active_ub_names, u_b_mult)
            i_min += n_act
            wrong_inds = where(u_b_mult < 0.0)[0]
            if wrong_inds.size > 0:
                names_neg = array(self.active_ub_names)[wrong_inds]
                LOGGER.warning(
                    "Negative Lagrange multipliers for "
                    "upper bounds on variables"
                    "%s !",
                    str(names_neg),
                )
        n_act = len(self.active_ineq_names)
        if n_act > 0:
            ineq_mult = multipliers[i_min : i_min + n_act]
            lag[self.INEQUALITY] = (self.active_ineq_names, ineq_mult)
            i_min += n_act
            wrong_inds = where(ineq_mult < 0.0)[0]
            if wrong_inds.size > 0:
                names_neg = array(self.active_ineq_names)[wrong_inds]
                LOGGER.warning(
                    "Negative Lagrange multipliers for "
                    "inequality constraints"
                    "%s !",
                    str(names_neg),
                )
        if self.active_eq_names:
            lag[self.EQUALITY] = (
                self.active_eq_names,
                multipliers[i_min : i_min + n_act],
            )
            i_min += n_act

        self.lagrange_multipliers = lag

    def _initialize_multipliers(self):
        """Initialize the Lagrange multipliers with zeros.

        :returns: Lagrange multipliers
        :rtype: dict(dict(float64))
        """
        problem = self.opt_problem
        multipliers = dict()

        # Bound-constraints
        indexed_varnames = problem.design_space.get_indexed_variables_names()
        multipliers[self.LOWER_BOUNDS] = dict.fromkeys(indexed_varnames, 0.0)
        multipliers[self.UPPER_BOUNDS] = dict.fromkeys(indexed_varnames, 0.0)

        # Inequality-constraints
        multipliers[self.INEQUALITY] = {
            func.name if func.dim == 1 else self._get_component_name(func.name, i): 0.0
            for func in problem.get_ineq_constraints()
            for i in range(func.dim)
        }

        # Equality-constraints
        multipliers[self.EQUALITY] = {
            func.name if func.dim == 1 else self._get_component_name(func.name, i): 0.0
            for func in problem.get_eq_constraints()
            for i in range(func.dim)
        }

        return multipliers

    def get_multipliers_arrays(self):
        """Return the Lagrange multipliers (zero and nonzero) as arrays.

        :returns: Lagrange multipliers
        :rtype: dict(dict(ndarray))
        """
        problem = self.opt_problem
        design_space = problem.design_space

        # Convert to dictionaries
        multipliers = dict()
        for label in self.CSTR_LABELS:
            names, mults = self.lagrange_multipliers.get(label, ([], array([])))
            multipliers[label] = dict(zip(names, mults))

        # Add the Lagrange multipliers equal to zero
        multipliers_init = self._initialize_multipliers()
        for label in self.CSTR_LABELS:
            multipliers_init[label].update(multipliers[label])

        # Cast the multipliers as arrays
        mult_arrays = dict()
        # Bound-constraints multipliers
        mult_arrays[self.LOWER_BOUNDS] = dict()
        mult_arrays[self.UPPER_BOUNDS] = dict()
        for name in design_space.variables_names:
            indexed_varnames = design_space.get_indexed_variables_names()
            var_low_mult = array(
                [
                    multipliers_init[self.LOWER_BOUNDS][comp_name]
                    for comp_name in indexed_varnames
                ]
            )
            mult_arrays[self.LOWER_BOUNDS][name] = var_low_mult
            var_upp_mult = array(
                [
                    multipliers_init[self.UPPER_BOUNDS][comp_name]
                    for comp_name in indexed_varnames
                ]
            )
            mult_arrays[self.UPPER_BOUNDS][name] = var_upp_mult
        # Inequality-constraints multipliers
        ineq_mult = multipliers_init[self.INEQUALITY]
        mult_arrays[self.INEQUALITY] = dict()
        for func in problem.get_ineq_constraints():
            func_mult = array(
                [
                    ineq_mult[
                        func.name
                        if func.dim == 1
                        else self._get_component_name(func.name, index)
                    ]
                    for index in range(func.dim)
                ]
            )
            mult_arrays[self.INEQUALITY][func.name] = func_mult
        # Equality-constraints multipliers
        eq_mult = multipliers_init[self.EQUALITY]
        mult_arrays[self.EQUALITY] = dict()
        for func in problem.get_eq_constraints():
            func_mult = array(
                [
                    eq_mult[
                        func.name
                        if func.dim == 1
                        else self._get_component_name(func.name, index)
                    ]
                    for index in range(func.dim)
                ]
            )
            mult_arrays[self.EQUALITY][func.name] = func_mult

        return mult_arrays

    @staticmethod
    def _get_component_name(name, index):
        """Return the name of a variable component.

        :param name: the variable name
        :type name: str
        :param index: the component index
        :type index: int
        :returns: name of the variable component
        :rtype: str
        """
        return name + DesignSpace.SEP + str(index)

    def _get_pretty_table(self):
        """Displays Lagrange Multipliers."""
        table = PrettyTable(
            ["Constraint type", "Constraint name", "Lagrange Multiplier"]
        )

        for cstr_type, nam_val in self.lagrange_multipliers.items():
            for name, value in zip(nam_val[0], nam_val[1]):
                table.add_row([cstr_type, name, value])

        return table

    def __str__(self, *args, **kwargs):
        """Textual representation of the design space."""
        desc = "Lagrange multipliers : "
        desc += "\n" + str(self._get_pretty_table().get_string())
        return desc
